"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class P4TypeInfo(google.protobuf.message.Message):
    """P4 type specs ---------------------------------------------------------------

    From the P4_16 spec:

    |--------------------|--------------------------------------------|
    |                    | Container type                             |
    | Element type       |-----------|--------------|-----------------|
    |                    | header    | header_union | struct or tuple |
    |--------------------|-----------|--------------|-----------------|
    | bit<W>             | allowed   | error        | allowed         |
    | int<W>             | allowed   | error        | allowed         |
    | varbit<W>          | allowed   | error        | allowed         |
    | int                | error     | error        | error           |
    | void               | error     | error        | error           |
    | error              | error     | error        | allowed         |
    | match_kind         | error     | error        | error           |
    | bool               | error     | error        | allowed         |
    | enum               | allowed*  | error        | allowed         |
    | header             | error     | allowed      | allowed         |
    | header stack       | error     | error        | allowed         |
    | header_union       | error     | error        | allowed         |
    | struct             | error     | error        | allowed         |
    | tuple              | error     | error        | allowed         |
    |--------------------|-----------|--------------|-----------------|

    *if serializable

    These P4 types (struct, header_type, header_union and enum) are guaranteed to
    have a fully-qualified name (e.g. you cannot use an anonymous struct to
    declare a variable like in C). Instead of duplicating the type spec for these
    every time the type is used, we include the type spec once in this P4TypeInfo
    message and refer to the types by name in the P4DataTypeSpec message. We also
    support annotations for these type specs which can be useful, e.g. to
    identify well-known headers (such as ipv4).
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class StructsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text
        @property
        def value(self) -> global___P4StructTypeSpec: ...
        def __init__(self,
            *,
            key: typing.Text = ...,
            value: typing.Optional[global___P4StructTypeSpec] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value",b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    class HeadersEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text
        @property
        def value(self) -> global___P4HeaderTypeSpec: ...
        def __init__(self,
            *,
            key: typing.Text = ...,
            value: typing.Optional[global___P4HeaderTypeSpec] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value",b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    class HeaderUnionsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text
        @property
        def value(self) -> global___P4HeaderUnionTypeSpec: ...
        def __init__(self,
            *,
            key: typing.Text = ...,
            value: typing.Optional[global___P4HeaderUnionTypeSpec] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value",b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    class EnumsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text
        @property
        def value(self) -> global___P4EnumTypeSpec: ...
        def __init__(self,
            *,
            key: typing.Text = ...,
            value: typing.Optional[global___P4EnumTypeSpec] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value",b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    class SerializableEnumsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text
        @property
        def value(self) -> global___P4SerializableEnumTypeSpec: ...
        def __init__(self,
            *,
            key: typing.Text = ...,
            value: typing.Optional[global___P4SerializableEnumTypeSpec] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value",b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    class NewTypesEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text
        @property
        def value(self) -> global___P4NewTypeSpec: ...
        def __init__(self,
            *,
            key: typing.Text = ...,
            value: typing.Optional[global___P4NewTypeSpec] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value",b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    STRUCTS_FIELD_NUMBER: builtins.int
    HEADERS_FIELD_NUMBER: builtins.int
    HEADER_UNIONS_FIELD_NUMBER: builtins.int
    ENUMS_FIELD_NUMBER: builtins.int
    ERROR_FIELD_NUMBER: builtins.int
    SERIALIZABLE_ENUMS_FIELD_NUMBER: builtins.int
    NEW_TYPES_FIELD_NUMBER: builtins.int
    @property
    def structs(self) -> google.protobuf.internal.containers.MessageMap[typing.Text, global___P4StructTypeSpec]: ...
    @property
    def headers(self) -> google.protobuf.internal.containers.MessageMap[typing.Text, global___P4HeaderTypeSpec]: ...
    @property
    def header_unions(self) -> google.protobuf.internal.containers.MessageMap[typing.Text, global___P4HeaderUnionTypeSpec]: ...
    @property
    def enums(self) -> google.protobuf.internal.containers.MessageMap[typing.Text, global___P4EnumTypeSpec]: ...
    @property
    def error(self) -> global___P4ErrorTypeSpec: ...
    @property
    def serializable_enums(self) -> google.protobuf.internal.containers.MessageMap[typing.Text, global___P4SerializableEnumTypeSpec]: ...
    @property
    def new_types(self) -> google.protobuf.internal.containers.MessageMap[typing.Text, global___P4NewTypeSpec]: ...
    def __init__(self,
        *,
        structs: typing.Optional[typing.Mapping[typing.Text, global___P4StructTypeSpec]] = ...,
        headers: typing.Optional[typing.Mapping[typing.Text, global___P4HeaderTypeSpec]] = ...,
        header_unions: typing.Optional[typing.Mapping[typing.Text, global___P4HeaderUnionTypeSpec]] = ...,
        enums: typing.Optional[typing.Mapping[typing.Text, global___P4EnumTypeSpec]] = ...,
        error: typing.Optional[global___P4ErrorTypeSpec] = ...,
        serializable_enums: typing.Optional[typing.Mapping[typing.Text, global___P4SerializableEnumTypeSpec]] = ...,
        new_types: typing.Optional[typing.Mapping[typing.Text, global___P4NewTypeSpec]] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["error",b"error"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["enums",b"enums","error",b"error","header_unions",b"header_unions","headers",b"headers","new_types",b"new_types","serializable_enums",b"serializable_enums","structs",b"structs"]) -> None: ...
global___P4TypeInfo = P4TypeInfo

class P4DataTypeSpec(google.protobuf.message.Message):
    """Describes a P4_16 type."""
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    BITSTRING_FIELD_NUMBER: builtins.int
    BOOL_FIELD_NUMBER: builtins.int
    TUPLE_FIELD_NUMBER: builtins.int
    STRUCT_FIELD_NUMBER: builtins.int
    HEADER_FIELD_NUMBER: builtins.int
    HEADER_UNION_FIELD_NUMBER: builtins.int
    HEADER_STACK_FIELD_NUMBER: builtins.int
    HEADER_UNION_STACK_FIELD_NUMBER: builtins.int
    ENUM_FIELD_NUMBER: builtins.int
    ERROR_FIELD_NUMBER: builtins.int
    SERIALIZABLE_ENUM_FIELD_NUMBER: builtins.int
    NEW_TYPE_FIELD_NUMBER: builtins.int
    @property
    def bitstring(self) -> global___P4BitstringLikeTypeSpec: ...
    @property
    def bool(self) -> global___P4BoolType: ...
    @property
    def tuple(self) -> global___P4TupleTypeSpec: ...
    @property
    def struct(self) -> global___P4NamedType: ...
    @property
    def header(self) -> global___P4NamedType: ...
    @property
    def header_union(self) -> global___P4NamedType: ...
    @property
    def header_stack(self) -> global___P4HeaderStackTypeSpec: ...
    @property
    def header_union_stack(self) -> global___P4HeaderUnionStackTypeSpec: ...
    @property
    def enum(self) -> global___P4NamedType: ...
    @property
    def error(self) -> global___P4ErrorType: ...
    @property
    def serializable_enum(self) -> global___P4NamedType: ...
    @property
    def new_type(self) -> global___P4NamedType: ...
    def __init__(self,
        *,
        bitstring: typing.Optional[global___P4BitstringLikeTypeSpec] = ...,
        bool: typing.Optional[global___P4BoolType] = ...,
        tuple: typing.Optional[global___P4TupleTypeSpec] = ...,
        struct: typing.Optional[global___P4NamedType] = ...,
        header: typing.Optional[global___P4NamedType] = ...,
        header_union: typing.Optional[global___P4NamedType] = ...,
        header_stack: typing.Optional[global___P4HeaderStackTypeSpec] = ...,
        header_union_stack: typing.Optional[global___P4HeaderUnionStackTypeSpec] = ...,
        enum: typing.Optional[global___P4NamedType] = ...,
        error: typing.Optional[global___P4ErrorType] = ...,
        serializable_enum: typing.Optional[global___P4NamedType] = ...,
        new_type: typing.Optional[global___P4NamedType] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["bitstring",b"bitstring","bool",b"bool","enum",b"enum","error",b"error","header",b"header","header_stack",b"header_stack","header_union",b"header_union","header_union_stack",b"header_union_stack","new_type",b"new_type","serializable_enum",b"serializable_enum","struct",b"struct","tuple",b"tuple","type_spec",b"type_spec"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["bitstring",b"bitstring","bool",b"bool","enum",b"enum","error",b"error","header",b"header","header_stack",b"header_stack","header_union",b"header_union","header_union_stack",b"header_union_stack","new_type",b"new_type","serializable_enum",b"serializable_enum","struct",b"struct","tuple",b"tuple","type_spec",b"type_spec"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["type_spec",b"type_spec"]) -> typing.Optional[typing_extensions.Literal["bitstring","bool","tuple","struct","header","header_union","header_stack","header_union_stack","enum","error","serializable_enum","new_type"]]: ...
global___P4DataTypeSpec = P4DataTypeSpec

class P4NamedType(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    NAME_FIELD_NUMBER: builtins.int
    name: typing.Text
    def __init__(self,
        *,
        name: typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["name",b"name"]) -> None: ...
global___P4NamedType = P4NamedType

class P4BoolType(google.protobuf.message.Message):
    """Empty message as no type information needed, just used as a placeholder in
    the oneof to identify boolean types.
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    def __init__(self,
        ) -> None: ...
global___P4BoolType = P4BoolType

class P4ErrorType(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    def __init__(self,
        ) -> None: ...
global___P4ErrorType = P4ErrorType

class P4BitstringLikeTypeSpec(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    BIT_FIELD_NUMBER: builtins.int
    INT_FIELD_NUMBER: builtins.int
    VARBIT_FIELD_NUMBER: builtins.int
    ANNOTATIONS_FIELD_NUMBER: builtins.int
    ANNOTATION_LOCATIONS_FIELD_NUMBER: builtins.int
    STRUCTURED_ANNOTATIONS_FIELD_NUMBER: builtins.int
    @property
    def bit(self) -> global___P4BitTypeSpec:
        """bit<W>"""
        pass
    @property
    def int(self) -> global___P4IntTypeSpec:
        """int<W>"""
        pass
    @property
    def varbit(self) -> global___P4VarbitTypeSpec:
        """varbit<W>"""
        pass
    @property
    def annotations(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]:
        """Useful to identify well-known types, such as IP address or Ethernet MAC
        address.
        """
        pass
    @property
    def annotation_locations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___SourceLocation]:
        """Optional. If present, the location of `annotations[i]` is given by
        `annotation_locations[i]`.
        """
        pass
    @property
    def structured_annotations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___StructuredAnnotation]: ...
    def __init__(self,
        *,
        bit: typing.Optional[global___P4BitTypeSpec] = ...,
        int: typing.Optional[global___P4IntTypeSpec] = ...,
        varbit: typing.Optional[global___P4VarbitTypeSpec] = ...,
        annotations: typing.Optional[typing.Iterable[typing.Text]] = ...,
        annotation_locations: typing.Optional[typing.Iterable[global___SourceLocation]] = ...,
        structured_annotations: typing.Optional[typing.Iterable[global___StructuredAnnotation]] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["bit",b"bit","int",b"int","type_spec",b"type_spec","varbit",b"varbit"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["annotation_locations",b"annotation_locations","annotations",b"annotations","bit",b"bit","int",b"int","structured_annotations",b"structured_annotations","type_spec",b"type_spec","varbit",b"varbit"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["type_spec",b"type_spec"]) -> typing.Optional[typing_extensions.Literal["bit","int","varbit"]]: ...
global___P4BitstringLikeTypeSpec = P4BitstringLikeTypeSpec

class P4BitTypeSpec(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    BITWIDTH_FIELD_NUMBER: builtins.int
    bitwidth: builtins.int
    def __init__(self,
        *,
        bitwidth: builtins.int = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["bitwidth",b"bitwidth"]) -> None: ...
global___P4BitTypeSpec = P4BitTypeSpec

class P4IntTypeSpec(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    BITWIDTH_FIELD_NUMBER: builtins.int
    bitwidth: builtins.int
    def __init__(self,
        *,
        bitwidth: builtins.int = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["bitwidth",b"bitwidth"]) -> None: ...
global___P4IntTypeSpec = P4IntTypeSpec

class P4VarbitTypeSpec(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    MAX_BITWIDTH_FIELD_NUMBER: builtins.int
    max_bitwidth: builtins.int
    def __init__(self,
        *,
        max_bitwidth: builtins.int = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["max_bitwidth",b"max_bitwidth"]) -> None: ...
global___P4VarbitTypeSpec = P4VarbitTypeSpec

class P4TupleTypeSpec(google.protobuf.message.Message):
    """From the P4_16 spec: "A tuple is similar to a struct, in that it holds
    multiple values. Unlike a struct type, tuples have no named fields."
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    MEMBERS_FIELD_NUMBER: builtins.int
    @property
    def members(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___P4DataTypeSpec]: ...
    def __init__(self,
        *,
        members: typing.Optional[typing.Iterable[global___P4DataTypeSpec]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["members",b"members"]) -> None: ...
global___P4TupleTypeSpec = P4TupleTypeSpec

class P4StructTypeSpec(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class Member(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        NAME_FIELD_NUMBER: builtins.int
        TYPE_SPEC_FIELD_NUMBER: builtins.int
        name: typing.Text
        @property
        def type_spec(self) -> global___P4DataTypeSpec: ...
        def __init__(self,
            *,
            name: typing.Text = ...,
            type_spec: typing.Optional[global___P4DataTypeSpec] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["type_spec",b"type_spec"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["name",b"name","type_spec",b"type_spec"]) -> None: ...

    MEMBERS_FIELD_NUMBER: builtins.int
    ANNOTATIONS_FIELD_NUMBER: builtins.int
    ANNOTATION_LOCATIONS_FIELD_NUMBER: builtins.int
    STRUCTURED_ANNOTATIONS_FIELD_NUMBER: builtins.int
    @property
    def members(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___P4StructTypeSpec.Member]: ...
    @property
    def annotations(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]: ...
    @property
    def annotation_locations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___SourceLocation]:
        """Optional. If present, the location of `annotations[i]` is given by
        `annotation_locations[i]`.
        """
        pass
    @property
    def structured_annotations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___StructuredAnnotation]: ...
    def __init__(self,
        *,
        members: typing.Optional[typing.Iterable[global___P4StructTypeSpec.Member]] = ...,
        annotations: typing.Optional[typing.Iterable[typing.Text]] = ...,
        annotation_locations: typing.Optional[typing.Iterable[global___SourceLocation]] = ...,
        structured_annotations: typing.Optional[typing.Iterable[global___StructuredAnnotation]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["annotation_locations",b"annotation_locations","annotations",b"annotations","members",b"members","structured_annotations",b"structured_annotations"]) -> None: ...
global___P4StructTypeSpec = P4StructTypeSpec

class P4HeaderTypeSpec(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class Member(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        NAME_FIELD_NUMBER: builtins.int
        TYPE_SPEC_FIELD_NUMBER: builtins.int
        name: typing.Text
        @property
        def type_spec(self) -> global___P4BitstringLikeTypeSpec: ...
        def __init__(self,
            *,
            name: typing.Text = ...,
            type_spec: typing.Optional[global___P4BitstringLikeTypeSpec] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["type_spec",b"type_spec"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["name",b"name","type_spec",b"type_spec"]) -> None: ...

    MEMBERS_FIELD_NUMBER: builtins.int
    ANNOTATIONS_FIELD_NUMBER: builtins.int
    ANNOTATION_LOCATIONS_FIELD_NUMBER: builtins.int
    STRUCTURED_ANNOTATIONS_FIELD_NUMBER: builtins.int
    @property
    def members(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___P4HeaderTypeSpec.Member]: ...
    @property
    def annotations(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]: ...
    @property
    def annotation_locations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___SourceLocation]:
        """Optional. If present, the location of `annotations[i]` is given by
        `annotation_locations[i]`.
        """
        pass
    @property
    def structured_annotations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___StructuredAnnotation]: ...
    def __init__(self,
        *,
        members: typing.Optional[typing.Iterable[global___P4HeaderTypeSpec.Member]] = ...,
        annotations: typing.Optional[typing.Iterable[typing.Text]] = ...,
        annotation_locations: typing.Optional[typing.Iterable[global___SourceLocation]] = ...,
        structured_annotations: typing.Optional[typing.Iterable[global___StructuredAnnotation]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["annotation_locations",b"annotation_locations","annotations",b"annotations","members",b"members","structured_annotations",b"structured_annotations"]) -> None: ...
global___P4HeaderTypeSpec = P4HeaderTypeSpec

class P4HeaderUnionTypeSpec(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class Member(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        NAME_FIELD_NUMBER: builtins.int
        HEADER_FIELD_NUMBER: builtins.int
        name: typing.Text
        @property
        def header(self) -> global___P4NamedType: ...
        def __init__(self,
            *,
            name: typing.Text = ...,
            header: typing.Optional[global___P4NamedType] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["header",b"header"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["header",b"header","name",b"name"]) -> None: ...

    MEMBERS_FIELD_NUMBER: builtins.int
    ANNOTATIONS_FIELD_NUMBER: builtins.int
    ANNOTATION_LOCATIONS_FIELD_NUMBER: builtins.int
    STRUCTURED_ANNOTATIONS_FIELD_NUMBER: builtins.int
    @property
    def members(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___P4HeaderUnionTypeSpec.Member]: ...
    @property
    def annotations(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]: ...
    @property
    def annotation_locations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___SourceLocation]:
        """Optional. If present, the location of `annotations[i]` is given by
        `annotation_locations[i]`.
        """
        pass
    @property
    def structured_annotations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___StructuredAnnotation]: ...
    def __init__(self,
        *,
        members: typing.Optional[typing.Iterable[global___P4HeaderUnionTypeSpec.Member]] = ...,
        annotations: typing.Optional[typing.Iterable[typing.Text]] = ...,
        annotation_locations: typing.Optional[typing.Iterable[global___SourceLocation]] = ...,
        structured_annotations: typing.Optional[typing.Iterable[global___StructuredAnnotation]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["annotation_locations",b"annotation_locations","annotations",b"annotations","members",b"members","structured_annotations",b"structured_annotations"]) -> None: ...
global___P4HeaderUnionTypeSpec = P4HeaderUnionTypeSpec

class P4HeaderStackTypeSpec(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    HEADER_FIELD_NUMBER: builtins.int
    SIZE_FIELD_NUMBER: builtins.int
    @property
    def header(self) -> global___P4NamedType: ...
    size: builtins.int
    def __init__(self,
        *,
        header: typing.Optional[global___P4NamedType] = ...,
        size: builtins.int = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["header",b"header"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["header",b"header","size",b"size"]) -> None: ...
global___P4HeaderStackTypeSpec = P4HeaderStackTypeSpec

class P4HeaderUnionStackTypeSpec(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    HEADER_UNION_FIELD_NUMBER: builtins.int
    SIZE_FIELD_NUMBER: builtins.int
    @property
    def header_union(self) -> global___P4NamedType: ...
    size: builtins.int
    def __init__(self,
        *,
        header_union: typing.Optional[global___P4NamedType] = ...,
        size: builtins.int = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["header_union",b"header_union"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["header_union",b"header_union","size",b"size"]) -> None: ...
global___P4HeaderUnionStackTypeSpec = P4HeaderUnionStackTypeSpec

class KeyValuePair(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    KEY_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    key: typing.Text
    @property
    def value(self) -> global___Expression: ...
    def __init__(self,
        *,
        key: typing.Text = ...,
        value: typing.Optional[global___Expression] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["value",b"value"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...
global___KeyValuePair = KeyValuePair

class KeyValuePairList(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    KV_PAIRS_FIELD_NUMBER: builtins.int
    @property
    def kv_pairs(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___KeyValuePair]: ...
    def __init__(self,
        *,
        kv_pairs: typing.Optional[typing.Iterable[global___KeyValuePair]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["kv_pairs",b"kv_pairs"]) -> None: ...
global___KeyValuePairList = KeyValuePairList

class Expression(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    STRING_VALUE_FIELD_NUMBER: builtins.int
    INT64_VALUE_FIELD_NUMBER: builtins.int
    BOOL_VALUE_FIELD_NUMBER: builtins.int
    string_value: typing.Text
    int64_value: builtins.int
    bool_value: builtins.bool
    def __init__(self,
        *,
        string_value: typing.Text = ...,
        int64_value: builtins.int = ...,
        bool_value: builtins.bool = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["bool_value",b"bool_value","int64_value",b"int64_value","string_value",b"string_value","value",b"value"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["bool_value",b"bool_value","int64_value",b"int64_value","string_value",b"string_value","value",b"value"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["value",b"value"]) -> typing.Optional[typing_extensions.Literal["string_value","int64_value","bool_value"]]: ...
global___Expression = Expression

class ExpressionList(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    EXPRESSIONS_FIELD_NUMBER: builtins.int
    @property
    def expressions(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Expression]: ...
    def __init__(self,
        *,
        expressions: typing.Optional[typing.Iterable[global___Expression]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["expressions",b"expressions"]) -> None: ...
global___ExpressionList = ExpressionList

class StructuredAnnotation(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    NAME_FIELD_NUMBER: builtins.int
    EXPRESSION_LIST_FIELD_NUMBER: builtins.int
    KV_PAIR_LIST_FIELD_NUMBER: builtins.int
    SOURCE_LOCATION_FIELD_NUMBER: builtins.int
    name: typing.Text
    @property
    def expression_list(self) -> global___ExpressionList: ...
    @property
    def kv_pair_list(self) -> global___KeyValuePairList: ...
    @property
    def source_location(self) -> global___SourceLocation:
        """Optional. Location of the '@' symbol of this annotation in the source code."""
        pass
    def __init__(self,
        *,
        name: typing.Text = ...,
        expression_list: typing.Optional[global___ExpressionList] = ...,
        kv_pair_list: typing.Optional[global___KeyValuePairList] = ...,
        source_location: typing.Optional[global___SourceLocation] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["body",b"body","expression_list",b"expression_list","kv_pair_list",b"kv_pair_list","source_location",b"source_location"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["body",b"body","expression_list",b"expression_list","kv_pair_list",b"kv_pair_list","name",b"name","source_location",b"source_location"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["body",b"body"]) -> typing.Optional[typing_extensions.Literal["expression_list","kv_pair_list"]]: ...
global___StructuredAnnotation = StructuredAnnotation

class SourceLocation(google.protobuf.message.Message):
    """Location of code relative to a given source file."""
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    FILE_FIELD_NUMBER: builtins.int
    LINE_FIELD_NUMBER: builtins.int
    COLUMN_FIELD_NUMBER: builtins.int
    file: typing.Text
    """Path to the source file (absolute or relative to the working directory)."""

    line: builtins.int
    """Line and column numbers within the source file, 1-based."""

    column: builtins.int
    def __init__(self,
        *,
        file: typing.Text = ...,
        line: builtins.int = ...,
        column: builtins.int = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["column",b"column","file",b"file","line",b"line"]) -> None: ...
global___SourceLocation = SourceLocation

class P4EnumTypeSpec(google.protobuf.message.Message):
    """For "safe" enums with no underlying representation and no member integer
    values.
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class Member(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        NAME_FIELD_NUMBER: builtins.int
        ANNOTATIONS_FIELD_NUMBER: builtins.int
        ANNOTATION_LOCATIONS_FIELD_NUMBER: builtins.int
        STRUCTURED_ANNOTATIONS_FIELD_NUMBER: builtins.int
        name: typing.Text
        @property
        def annotations(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]: ...
        @property
        def annotation_locations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___SourceLocation]:
            """Optional. If present, the location of `annotations[i]` is given by
            `annotation_locations[i]`.
            """
            pass
        @property
        def structured_annotations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___StructuredAnnotation]: ...
        def __init__(self,
            *,
            name: typing.Text = ...,
            annotations: typing.Optional[typing.Iterable[typing.Text]] = ...,
            annotation_locations: typing.Optional[typing.Iterable[global___SourceLocation]] = ...,
            structured_annotations: typing.Optional[typing.Iterable[global___StructuredAnnotation]] = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["annotation_locations",b"annotation_locations","annotations",b"annotations","name",b"name","structured_annotations",b"structured_annotations"]) -> None: ...

    MEMBERS_FIELD_NUMBER: builtins.int
    ANNOTATIONS_FIELD_NUMBER: builtins.int
    ANNOTATION_LOCATIONS_FIELD_NUMBER: builtins.int
    STRUCTURED_ANNOTATIONS_FIELD_NUMBER: builtins.int
    @property
    def members(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___P4EnumTypeSpec.Member]: ...
    @property
    def annotations(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]: ...
    @property
    def annotation_locations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___SourceLocation]:
        """Optional. If present, the location of `annotations[i]` is given by
        `annotation_locations[i]`.
        """
        pass
    @property
    def structured_annotations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___StructuredAnnotation]: ...
    def __init__(self,
        *,
        members: typing.Optional[typing.Iterable[global___P4EnumTypeSpec.Member]] = ...,
        annotations: typing.Optional[typing.Iterable[typing.Text]] = ...,
        annotation_locations: typing.Optional[typing.Iterable[global___SourceLocation]] = ...,
        structured_annotations: typing.Optional[typing.Iterable[global___StructuredAnnotation]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["annotation_locations",b"annotation_locations","annotations",b"annotations","members",b"members","structured_annotations",b"structured_annotations"]) -> None: ...
global___P4EnumTypeSpec = P4EnumTypeSpec

class P4SerializableEnumTypeSpec(google.protobuf.message.Message):
    """For serializable (or "unsafe") enums, which have an underlying type. Note
    that as per the P4_16 specification, the underlying representation can only
    be a bit<W> type.
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class Member(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        NAME_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        ANNOTATIONS_FIELD_NUMBER: builtins.int
        ANNOTATION_LOCATIONS_FIELD_NUMBER: builtins.int
        STRUCTURED_ANNOTATIONS_FIELD_NUMBER: builtins.int
        name: typing.Text
        value: builtins.bytes
        @property
        def annotations(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]: ...
        @property
        def annotation_locations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___SourceLocation]:
            """Optional. If present, the location of `annotations[i]` is given by
            `annotation_locations[i]`.
            """
            pass
        @property
        def structured_annotations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___StructuredAnnotation]: ...
        def __init__(self,
            *,
            name: typing.Text = ...,
            value: builtins.bytes = ...,
            annotations: typing.Optional[typing.Iterable[typing.Text]] = ...,
            annotation_locations: typing.Optional[typing.Iterable[global___SourceLocation]] = ...,
            structured_annotations: typing.Optional[typing.Iterable[global___StructuredAnnotation]] = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["annotation_locations",b"annotation_locations","annotations",b"annotations","name",b"name","structured_annotations",b"structured_annotations","value",b"value"]) -> None: ...

    UNDERLYING_TYPE_FIELD_NUMBER: builtins.int
    MEMBERS_FIELD_NUMBER: builtins.int
    ANNOTATIONS_FIELD_NUMBER: builtins.int
    ANNOTATION_LOCATIONS_FIELD_NUMBER: builtins.int
    STRUCTURED_ANNOTATIONS_FIELD_NUMBER: builtins.int
    @property
    def underlying_type(self) -> global___P4BitTypeSpec: ...
    @property
    def members(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___P4SerializableEnumTypeSpec.Member]: ...
    @property
    def annotations(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]: ...
    @property
    def annotation_locations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___SourceLocation]:
        """Optional. If present, the location of `annotations[i]` is given by
        `annotation_locations[i]`.
        """
        pass
    @property
    def structured_annotations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___StructuredAnnotation]: ...
    def __init__(self,
        *,
        underlying_type: typing.Optional[global___P4BitTypeSpec] = ...,
        members: typing.Optional[typing.Iterable[global___P4SerializableEnumTypeSpec.Member]] = ...,
        annotations: typing.Optional[typing.Iterable[typing.Text]] = ...,
        annotation_locations: typing.Optional[typing.Iterable[global___SourceLocation]] = ...,
        structured_annotations: typing.Optional[typing.Iterable[global___StructuredAnnotation]] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["underlying_type",b"underlying_type"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["annotation_locations",b"annotation_locations","annotations",b"annotations","members",b"members","structured_annotations",b"structured_annotations","underlying_type",b"underlying_type"]) -> None: ...
global___P4SerializableEnumTypeSpec = P4SerializableEnumTypeSpec

class P4ErrorTypeSpec(google.protobuf.message.Message):
    """Similar to an enum, but there is always one and only one instance per P4
    program.
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    MEMBERS_FIELD_NUMBER: builtins.int
    @property
    def members(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]: ...
    def __init__(self,
        *,
        members: typing.Optional[typing.Iterable[typing.Text]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["members",b"members"]) -> None: ...
global___P4ErrorTypeSpec = P4ErrorTypeSpec

class P4NewTypeTranslation(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class SdnString(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        def __init__(self,
            ) -> None: ...

    URI_FIELD_NUMBER: builtins.int
    SDN_BITWIDTH_FIELD_NUMBER: builtins.int
    SDN_STRING_FIELD_NUMBER: builtins.int
    uri: typing.Text
    """the URI uniquely identifies the translation in order to enable the
    P4Runtime agent to perform value-mapping appropriately when required. It is
    recommended that the URI includes at least the P4 architecture name and the
    type name.
    """

    sdn_bitwidth: builtins.int
    @property
    def sdn_string(self) -> global___P4NewTypeTranslation.SdnString: ...
    def __init__(self,
        *,
        uri: typing.Text = ...,
        sdn_bitwidth: builtins.int = ...,
        sdn_string: typing.Optional[global___P4NewTypeTranslation.SdnString] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["sdn_bitwidth",b"sdn_bitwidth","sdn_string",b"sdn_string","sdn_type",b"sdn_type"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["sdn_bitwidth",b"sdn_bitwidth","sdn_string",b"sdn_string","sdn_type",b"sdn_type","uri",b"uri"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["sdn_type",b"sdn_type"]) -> typing.Optional[typing_extensions.Literal["sdn_bitwidth","sdn_string"]]: ...
global___P4NewTypeTranslation = P4NewTypeTranslation

class P4NewTypeSpec(google.protobuf.message.Message):
    """New types introduced with the "type" keyword"""
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    ORIGINAL_TYPE_FIELD_NUMBER: builtins.int
    TRANSLATED_TYPE_FIELD_NUMBER: builtins.int
    ANNOTATIONS_FIELD_NUMBER: builtins.int
    ANNOTATION_LOCATIONS_FIELD_NUMBER: builtins.int
    STRUCTURED_ANNOTATIONS_FIELD_NUMBER: builtins.int
    @property
    def original_type(self) -> global___P4DataTypeSpec:
        """if no @p4runtime_translation annotation present"""
        pass
    @property
    def translated_type(self) -> global___P4NewTypeTranslation:
        """if @p4runtime_translation annotation present"""
        pass
    @property
    def annotations(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]:
        """for other annotations (not @p4runtime_translation)"""
        pass
    @property
    def annotation_locations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___SourceLocation]:
        """Optional. If present, the location of `annotations[i]` is given by
        `annotation_locations[i]`.
        """
        pass
    @property
    def structured_annotations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___StructuredAnnotation]: ...
    def __init__(self,
        *,
        original_type: typing.Optional[global___P4DataTypeSpec] = ...,
        translated_type: typing.Optional[global___P4NewTypeTranslation] = ...,
        annotations: typing.Optional[typing.Iterable[typing.Text]] = ...,
        annotation_locations: typing.Optional[typing.Iterable[global___SourceLocation]] = ...,
        structured_annotations: typing.Optional[typing.Iterable[global___StructuredAnnotation]] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["original_type",b"original_type","representation",b"representation","translated_type",b"translated_type"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["annotation_locations",b"annotation_locations","annotations",b"annotations","original_type",b"original_type","representation",b"representation","structured_annotations",b"structured_annotations","translated_type",b"translated_type"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["representation",b"representation"]) -> typing.Optional[typing_extensions.Literal["original_type","translated_type"]]: ...
global___P4NewTypeSpec = P4NewTypeSpec



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: p4/v1/p4data.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12p4/v1/p4data.proto\x12\x05p4.v1\"\x94\x03\n\x06P4Data\x12\x13\n\tbitstring\x18\x01 \x01(\x0cH\x00\x12!\n\x06varbit\x18\x02 \x01(\x0b\x32\x0f.p4.v1.P4VarbitH\x00\x12\x0e\n\x04\x62ool\x18\x03 \x01(\x08H\x00\x12$\n\x05tuple\x18\x04 \x01(\x0b\x32\x13.p4.v1.P4StructLikeH\x00\x12%\n\x06struct\x18\x05 \x01(\x0b\x32\x13.p4.v1.P4StructLikeH\x00\x12!\n\x06header\x18\x06 \x01(\x0b\x32\x0f.p4.v1.P4HeaderH\x00\x12,\n\x0cheader_union\x18\x07 \x01(\x0b\x32\x14.p4.v1.P4HeaderUnionH\x00\x12,\n\x0cheader_stack\x18\x08 \x01(\x0b\x32\x14.p4.v1.P4HeaderStackH\x00\x12\x37\n\x12header_union_stack\x18\t \x01(\x0b\x32\x19.p4.v1.P4HeaderUnionStackH\x00\x12\x0e\n\x04\x65num\x18\n \x01(\tH\x00\x12\x0f\n\x05\x65rror\x18\x0b \x01(\tH\x00\x12\x14\n\nenum_value\x18\x0c \x01(\x0cH\x00\x42\x06\n\x04\x64\x61ta\"/\n\x08P4Varbit\x12\x11\n\tbitstring\x18\x01 \x01(\x0c\x12\x10\n\x08\x62itwidth\x18\x02 \x01(\x05\".\n\x0cP4StructLike\x12\x1e\n\x07members\x18\x01 \x03(\x0b\x32\r.p4.v1.P4Data\"0\n\x08P4Header\x12\x10\n\x08is_valid\x18\x01 \x01(\x08\x12\x12\n\nbitstrings\x18\x02 \x03(\x0c\"Q\n\rP4HeaderUnion\x12\x19\n\x11valid_header_name\x18\x01 \x01(\t\x12%\n\x0cvalid_header\x18\x02 \x01(\x0b\x32\x0f.p4.v1.P4Header\"1\n\rP4HeaderStack\x12 \n\x07\x65ntries\x18\x01 \x03(\x0b\x32\x0f.p4.v1.P4Header\";\n\x12P4HeaderUnionStack\x12%\n\x07\x65ntries\x18\x01 \x03(\x0b\x32\x14.p4.v1.P4HeaderUnionB&Z$github.com/p4lang/p4runtime/go/p4/v1b\x06proto3')



_P4DATA = DESCRIPTOR.message_types_by_name['P4Data']
_P4VARBIT = DESCRIPTOR.message_types_by_name['P4Varbit']
_P4STRUCTLIKE = DESCRIPTOR.message_types_by_name['P4StructLike']
_P4HEADER = DESCRIPTOR.message_types_by_name['P4Header']
_P4HEADERUNION = DESCRIPTOR.message_types_by_name['P4HeaderUnion']
_P4HEADERSTACK = DESCRIPTOR.message_types_by_name['P4HeaderStack']
_P4HEADERUNIONSTACK = DESCRIPTOR.message_types_by_name['P4HeaderUnionStack']
P4Data = _reflection.GeneratedProtocolMessageType('P4Data', (_message.Message,), {
  'DESCRIPTOR' : _P4DATA,
  '__module__' : 'p4.v1.p4data_pb2'
  # @@protoc_insertion_point(class_scope:p4.v1.P4Data)
  })
_sym_db.RegisterMessage(P4Data)

P4Varbit = _reflection.GeneratedProtocolMessageType('P4Varbit', (_message.Message,), {
  'DESCRIPTOR' : _P4VARBIT,
  '__module__' : 'p4.v1.p4data_pb2'
  # @@protoc_insertion_point(class_scope:p4.v1.P4Varbit)
  })
_sym_db.RegisterMessage(P4Varbit)

P4StructLike = _reflection.GeneratedProtocolMessageType('P4StructLike', (_message.Message,), {
  'DESCRIPTOR' : _P4STRUCTLIKE,
  '__module__' : 'p4.v1.p4data_pb2'
  # @@protoc_insertion_point(class_scope:p4.v1.P4StructLike)
  })
_sym_db.RegisterMessage(P4StructLike)

P4Header = _reflection.GeneratedProtocolMessageType('P4Header', (_message.Message,), {
  'DESCRIPTOR' : _P4HEADER,
  '__module__' : 'p4.v1.p4data_pb2'
  # @@protoc_insertion_point(class_scope:p4.v1.P4Header)
  })
_sym_db.RegisterMessage(P4Header)

P4HeaderUnion = _reflection.GeneratedProtocolMessageType('P4HeaderUnion', (_message.Message,), {
  'DESCRIPTOR' : _P4HEADERUNION,
  '__module__' : 'p4.v1.p4data_pb2'
  # @@protoc_insertion_point(class_scope:p4.v1.P4HeaderUnion)
  })
_sym_db.RegisterMessage(P4HeaderUnion)

P4HeaderStack = _reflection.GeneratedProtocolMessageType('P4HeaderStack', (_message.Message,), {
  'DESCRIPTOR' : _P4HEADERSTACK,
  '__module__' : 'p4.v1.p4data_pb2'
  # @@protoc_insertion_point(class_scope:p4.v1.P4HeaderStack)
  })
_sym_db.RegisterMessage(P4HeaderStack)

P4HeaderUnionStack = _reflection.GeneratedProtocolMessageType('P4HeaderUnionStack', (_message.Message,), {
  'DESCRIPTOR' : _P4HEADERUNIONSTACK,
  '__module__' : 'p4.v1.p4data_pb2'
  # @@protoc_insertion_point(class_scope:p4.v1.P4HeaderUnionStack)
  })
_sym_db.RegisterMessage(P4HeaderUnionStack)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z$github.com/p4lang/p4runtime/go/p4/v1'
  _P4DATA._serialized_start=30
  _P4DATA._serialized_end=434
  _P4VARBIT._serialized_start=436
  _P4VARBIT._serialized_end=483
  _P4STRUCTLIKE._serialized_start=485
  _P4STRUCTLIKE._serialized_end=531
  _P4HEADER._serialized_start=533
  _P4HEADER._serialized_end=581
  _P4HEADERUNION._serialized_start=583
  _P4HEADERUNION._serialized_end=664
  _P4HEADERSTACK._serialized_start=666
  _P4HEADERSTACK._serialized_end=715
  _P4HEADERUNIONSTACK._serialized_start=717
  _P4HEADERUNIONSTACK._serialized_end=776
# @@protoc_insertion_point(module_scope)



"Logging utilities."

# Copyright (c) 2022 Bill Fisher
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import functools
import logging
import os
import sys
from typing import TYPE_CHECKING, Any, MutableMapping, TypeAlias


def get_setting(name: str, default: str = "") -> bool:
    "Retrieve environment variable setting."
    value = os.environ.get(name, default)
    if not value:
        return False
    return value.strip().lower() not in {"0", "false"}


FINSY_DEBUG = get_setting("FINSY_DEBUG")
FINSY_TRANSLATE_LOGS = get_setting("FINSY_TRANSLATE_LOGS", "true")


if TYPE_CHECKING:
    _LoggerAdapter: TypeAlias = logging.LoggerAdapter[logging.Logger]
else:
    # LoggerAdapter will be generic at runtime in Python 3.11.
    _LoggerAdapter: TypeAlias = logging.LoggerAdapter


class _CustomAdapter(_LoggerAdapter):
    """Custom log adapter to include the name of the current task."""

    def process(
        self,
        msg: Any,
        kwargs: MutableMapping[str, Any],
    ) -> tuple[Any, MutableMapping[str, Any]]:
        """Process the logging message and keyword arguments passed in to a
        logging call to insert contextual information.
        """

        task_name = ""
        try:
            # current_task() will raise a RuntimeError if there is no running
            # event loop. It can also return None if there's a running event
            # loop but we aren't in a task (ie we're in a low-level callback).
            task = asyncio.current_task()
            if task:
                task_name = task.get_name()
                # Remove the "fy:" prefix from the task name before logging.
                if task_name.startswith("fy:"):
                    task_name = task_name[3:]
                task_name = f"[{task_name}] "
            else:
                task_name = "[] "  # low-level callback - no running task
        except RuntimeError:
            pass

        return f"{task_name}{msg}", kwargs


LOGGER = _CustomAdapter(logging.getLogger(__package__))
MSG_LOG = _CustomAdapter(logging.getLogger(f"{__package__}.msg"))
TRACE_LOG = _CustomAdapter(logging.getLogger(f"{__package__}.trace"))


def _exc() -> BaseException | None:
    # TODO: replace sys.exc_info() with sys.exception() someday...
    return sys.exc_info()[1]


def _trace_step(func: Any) -> Any:
    "Decorator to log when we step in and step out of a coroutine function."

    @functools.wraps(func)
    async def _wrapper(*args: Any, **kwd: Any):
        try:
            TRACE_LOG.info("%s stepin", func.__qualname__)
            return await func(*args, **kwd)
        finally:
            TRACE_LOG.info("%s stepout ex=%r", func.__qualname__, _exc())

    return _wrapper


def _trace_noop(func: Any) -> Any:
    "No op decorator"
    return func


TRACE = _trace_step if FINSY_DEBUG else _trace_noop



"""Package that provides accessors for protobuf files.

For example, to access `p4.v1.p4runtime_pb2.CapabilitiesRequest`:

    from finsy.proto import p4r
    msg = p4r.CapabilitiesRequest(...)

Here is the full table:

    p4.v1.p4runtime_pb2       => `from finsy.proto import p4r`
    p4.v1.p4runtime_pb2_grpc  => `from finsy.proto import p4r_gprc`
    p4.v1.p4data_pb2          => `from finsy.proto import p4d`
    p4.config.v1.p4info_pb2   => `from finsy.proto import p4i`
    p4.config.v1.p4types_pb2  => `from finsy.proto import p4t`
    google.rpc.code_pb2       => `from finsy.proto import rpc_code`
    google.rpc.status_pb2     => `from finsy.proto import rpc_status`
    gnmi1.gnmi_pb2            => `from finsy.proto import gnmi`
    gnmi1.gnmi_pb2_grpc       => `from finsy.proto import gnmi_grpc`
    gnmi1.gnmi_ext_pb2        => `from finsy.proto import gnmi_ext`

"""

__all__ = (
    "gnmi",
    "gnmi_ext",
    "gnmi_grpc",
    "p4i",
    "p4t",
    "p4d",
    "p4r",
    "p4r_grpc",
    "rpc_code",
    "rpc_status",
    "U128",
)

import os
import sys

# Generated protobuf files reference their dependencies using absolute imports.
# To work around this, modify the import path to substitute our embedded
# modules for global "p4" and "google.rpc" modules. After we've imported the
# protobuf modules, we switch `sys.path` back.

sys.path.insert(0, os.path.dirname(__file__))

from gnmi1 import gnmi_ext_pb2 as gnmi_ext
from gnmi1 import gnmi_pb2 as gnmi
from gnmi1 import gnmi_pb2_grpc as gnmi_grpc
from google.rpc import code_pb2 as rpc_code
from google.rpc import status_pb2 as rpc_status
from p4.config.v1 import p4info_pb2 as p4i
from p4.config.v1 import p4types_pb2 as p4t
from p4.v1 import p4data_pb2 as p4d
from p4.v1 import p4runtime_pb2 as p4r
from p4.v1 import p4runtime_pb2_grpc as p4r_grpc

del sys.path[0]


class U128:
    "Utility class for p4r.Uint128."

    @staticmethod
    def encode(value: int) -> p4r.Uint128:
        "Create a Uint128 object from an integer."

        if not isinstance(value, int) or value < 0:
            raise ValueError(f"invalid argument: {value!r}")
        return p4r.Uint128(high=value >> 64, low=value & 0xFFFFFFFFFFFFFFFF)

    @staticmethod
    def decode(value: p4r.Uint128) -> int:
        "Convert a Uint128 object to an integer."

        return (value.high << 64) | value.low



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class P4Data(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    BITSTRING_FIELD_NUMBER: builtins.int
    VARBIT_FIELD_NUMBER: builtins.int
    BOOL_FIELD_NUMBER: builtins.int
    TUPLE_FIELD_NUMBER: builtins.int
    STRUCT_FIELD_NUMBER: builtins.int
    HEADER_FIELD_NUMBER: builtins.int
    HEADER_UNION_FIELD_NUMBER: builtins.int
    HEADER_STACK_FIELD_NUMBER: builtins.int
    HEADER_UNION_STACK_FIELD_NUMBER: builtins.int
    ENUM_FIELD_NUMBER: builtins.int
    ERROR_FIELD_NUMBER: builtins.int
    ENUM_VALUE_FIELD_NUMBER: builtins.int
    bitstring: builtins.bytes
    """for bit<W>, int<W>"""

    @property
    def varbit(self) -> global___P4Varbit:
        """for varbit<W>"""
        pass
    bool: builtins.bool
    @property
    def tuple(self) -> global___P4StructLike: ...
    @property
    def struct(self) -> global___P4StructLike: ...
    @property
    def header(self) -> global___P4Header: ...
    @property
    def header_union(self) -> global___P4HeaderUnion: ...
    @property
    def header_stack(self) -> global___P4HeaderStack: ...
    @property
    def header_union_stack(self) -> global___P4HeaderUnionStack: ...
    enum: typing.Text
    """safe (non-serializable) enums only"""

    error: typing.Text
    enum_value: builtins.bytes
    """serializable enums only"""

    def __init__(self,
        *,
        bitstring: builtins.bytes = ...,
        varbit: typing.Optional[global___P4Varbit] = ...,
        bool: builtins.bool = ...,
        tuple: typing.Optional[global___P4StructLike] = ...,
        struct: typing.Optional[global___P4StructLike] = ...,
        header: typing.Optional[global___P4Header] = ...,
        header_union: typing.Optional[global___P4HeaderUnion] = ...,
        header_stack: typing.Optional[global___P4HeaderStack] = ...,
        header_union_stack: typing.Optional[global___P4HeaderUnionStack] = ...,
        enum: typing.Text = ...,
        error: typing.Text = ...,
        enum_value: builtins.bytes = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["bitstring",b"bitstring","bool",b"bool","data",b"data","enum",b"enum","enum_value",b"enum_value","error",b"error","header",b"header","header_stack",b"header_stack","header_union",b"header_union","header_union_stack",b"header_union_stack","struct",b"struct","tuple",b"tuple","varbit",b"varbit"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["bitstring",b"bitstring","bool",b"bool","data",b"data","enum",b"enum","enum_value",b"enum_value","error",b"error","header",b"header","header_stack",b"header_stack","header_union",b"header_union","header_union_stack",b"header_union_stack","struct",b"struct","tuple",b"tuple","varbit",b"varbit"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["data",b"data"]) -> typing.Optional[typing_extensions.Literal["bitstring","varbit","bool","tuple","struct","header","header_union","header_stack","header_union_stack","enum","error","enum_value"]]: ...
global___P4Data = P4Data

class P4Varbit(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    BITSTRING_FIELD_NUMBER: builtins.int
    BITWIDTH_FIELD_NUMBER: builtins.int
    bitstring: builtins.bytes
    bitwidth: builtins.int
    """dynamic bitwidth of the field"""

    def __init__(self,
        *,
        bitstring: builtins.bytes = ...,
        bitwidth: builtins.int = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["bitstring",b"bitstring","bitwidth",b"bitwidth"]) -> None: ...
global___P4Varbit = P4Varbit

class P4StructLike(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    MEMBERS_FIELD_NUMBER: builtins.int
    @property
    def members(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___P4Data]: ...
    def __init__(self,
        *,
        members: typing.Optional[typing.Iterable[global___P4Data]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["members",b"members"]) -> None: ...
global___P4StructLike = P4StructLike

class P4Header(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    IS_VALID_FIELD_NUMBER: builtins.int
    BITSTRINGS_FIELD_NUMBER: builtins.int
    is_valid: builtins.bool
    """If the header is invalid (is_valid is "false"), then the bitstrings
    repeated field must be empty.
    """

    @property
    def bitstrings(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bytes]: ...
    def __init__(self,
        *,
        is_valid: builtins.bool = ...,
        bitstrings: typing.Optional[typing.Iterable[builtins.bytes]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["bitstrings",b"bitstrings","is_valid",b"is_valid"]) -> None: ...
global___P4Header = P4Header

class P4HeaderUnion(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    VALID_HEADER_NAME_FIELD_NUMBER: builtins.int
    VALID_HEADER_FIELD_NUMBER: builtins.int
    valid_header_name: typing.Text
    """An empty string indicates that none of the union members are valid and
    valid_header must therefore be unset.
    """

    @property
    def valid_header(self) -> global___P4Header: ...
    def __init__(self,
        *,
        valid_header_name: typing.Text = ...,
        valid_header: typing.Optional[global___P4Header] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["valid_header",b"valid_header"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["valid_header",b"valid_header","valid_header_name",b"valid_header_name"]) -> None: ...
global___P4HeaderUnion = P4HeaderUnion

class P4HeaderStack(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    ENTRIES_FIELD_NUMBER: builtins.int
    @property
    def entries(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___P4Header]:
        """The length of this repeated field must always be equal to the compile-time
        size of the header stack, which is specified in P4Info.
        """
        pass
    def __init__(self,
        *,
        entries: typing.Optional[typing.Iterable[global___P4Header]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["entries",b"entries"]) -> None: ...
global___P4HeaderStack = P4HeaderStack

class P4HeaderUnionStack(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    ENTRIES_FIELD_NUMBER: builtins.int
    @property
    def entries(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___P4HeaderUnion]:
        """The length of this repeated field must always be equal to the compile-time
        size of the header union stack, which is specified in P4Info.
        """
        pass
    def __init__(self,
        *,
        entries: typing.Optional[typing.Iterable[global___P4HeaderUnion]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["entries",b"entries"]) -> None: ...
global___P4HeaderUnionStack = P4HeaderUnionStack



"Implements entity encode/decode functions."

# Copyright (c) 2022 Bill Fisher
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections.abc
from dataclasses import KW_ONLY, dataclass
from typing import Any, Iterator, NoReturn, Protocol, Sequence

from typing_extensions import Self

from finsy.log import LOGGER
from finsy.p4schema import P4Action, P4ActionRef, P4Schema, P4Table, P4UpdateType
from finsy.proto import p4r


class _SupportsDecode(Protocol):
    # FIXME: This is a class method that returns an instance of that class.
    def decode(self, msg: Any, schema: P4Schema) -> Any:
        ...


class _SupportsEncodeEntity(Protocol):
    def encode(self, schema: P4Schema) -> p4r.Entity:
        ...


class _SupportsEncodeUpdate(Protocol):
    def encode_update(self, schema: P4Schema) -> p4r.Update | p4r.StreamMessageRequest:
        ...


_DECODER: dict[str, _SupportsDecode] = {}


def decodable(key: str):
    "Class decorator to specify the class used to decode P4Runtime messages."

    def _decorate(cls: Any):
        assert key not in _DECODER
        _DECODER[key] = cls
        return cls

    return _decorate


def decode_entity(msg: p4r.Entity, schema: P4Schema) -> Any:
    "Decode a P4Runtime Entity to the Python class registered in _DECODER."

    key = msg.WhichOneof("entity")
    if key is None:
        raise ValueError("missing entity")

    if key == "packet_replication_engine_entry":
        submsg = msg.packet_replication_engine_entry
        key = submsg.WhichOneof("type")
        if key is None:
            raise ValueError("missing type")

    return _DECODER[key].decode(msg, schema)


def decode_stream(msg: p4r.StreamMessageResponse, schema: P4Schema) -> Any:
    "Decode a StreamMessageResponse to the class registered in _DECODER."

    key = msg.WhichOneof("update")
    if key is None:
        raise ValueError("missing update")

    return _DECODER[key].decode(msg, schema)


# Recursive typedefs.
EntityList = p4r.Entity | _SupportsEncodeEntity | Sequence["EntityList"]
UpdateList = (
    p4r.Update
    | p4r.StreamMessageRequest
    | _SupportsEncodeUpdate
    | Sequence["UpdateList"]
)


def _flatten(values: Any) -> Iterator[Any]:
    "Flatten lists and tuples."
    for val in values:
        if isinstance(val, collections.abc.Sequence):
            yield from _flatten(val)
        else:
            yield val


def _encode_entity(
    value: p4r.Entity | _SupportsEncodeEntity,
    schema: P4Schema,
) -> p4r.Entity:
    "Encode an entity, if necessary."

    if isinstance(value, p4r.Entity):
        return value

    return value.encode(schema)


def encode_entities(values: EntityList, schema: P4Schema) -> list[p4r.Entity]:
    """Convert list of python objects to list of P4Runtime Entities."""

    if not isinstance(values, collections.abc.Sequence):
        return [_encode_entity(values, schema)]

    return [_encode_entity(val, schema) for val in _flatten(values)]


def _encode_update(
    value: p4r.Update | p4r.StreamMessageRequest | _SupportsEncodeUpdate,
    schema: P4Schema,
) -> p4r.Update | p4r.StreamMessageRequest:
    "Encode an Update or outgoing message request, if necessary."

    if isinstance(value, (p4r.Update, p4r.StreamMessageRequest)):
        return value

    return value.encode_update(schema)


def encode_updates(
    values: UpdateList,
    schema: P4Schema,
) -> list[p4r.Update | p4r.StreamMessageRequest]:
    """Convert list of python objects to P4Runtime Updates or request messages."""

    if not isinstance(values, collections.abc.Sequence):
        return [_encode_update(values, schema)]

    return [_encode_update(val, schema) for val in _flatten(values)]


class _Writable:
    _update_type: P4UpdateType = P4UpdateType.UNSPECIFIED

    def __pos__(self):
        self._update_type = P4UpdateType.INSERT
        return self

    def __neg__(self):
        self._update_type = P4UpdateType.DELETE
        return self

    def __invert__(self):
        self._update_type = P4UpdateType.MODIFY
        return self

    def encode(self, _schema: P4Schema) -> p4r.Entity:
        raise NotImplementedError()

    def encode_update(self, schema: P4Schema) -> p4r.Update:
        if self._update_type == P4UpdateType.UNSPECIFIED:
            raise ValueError("unspecified update type")
        return p4r.Update(type=self._update_type.vt(), entity=self.encode(schema))


_DataValueType = Any  # TODO: tighten P4Data type constraint later
_MetadataDictType = dict[str, Any]
_PortType = int
_ReplicaCanonType = tuple[_PortType, int]
_ReplicaType = _ReplicaCanonType | _PortType
_Weight = int | tuple[int, int | bytes]


def encode_replica(value: _ReplicaType) -> p4r.Replica:
    "Convert python representation to Replica."

    if isinstance(value, int):
        egress_port, instance = value, 0
    else:
        assert isinstance(value, tuple) and len(value) == 2
        egress_port, instance = value

    return p4r.Replica(egress_port=egress_port, instance=instance)


def decode_replica(replica: p4r.Replica) -> _ReplicaCanonType:
    "Convert Replica to python representation."

    return (replica.egress_port, replica.instance)


class P4TableMatch(dict[str, Any]):
    "Represents a sequence of P4Runtime FieldMatch."

    def encode(self, table: P4Table) -> list[p4r.FieldMatch]:
        "Encode TableMatch data as protobuf."
        result: list[p4r.FieldMatch] = []
        match_fields = table.match_fields

        for key, value in self.items():
            try:
                field = match_fields[key].encode(value)
                if field is not None:
                    result.append(field)
            except Exception as ex:
                raise ValueError(f"{table.name!r}: Match field {key!r}: {ex}") from ex

        return result

    @classmethod
    def decode(cls, msgs: Sequence[p4r.FieldMatch], table: P4Table) -> Self:
        "Decode protobuf to TableMatch data."
        result = {}
        match_fields = table.match_fields

        for field in msgs:
            fld = match_fields[field.field_id]
            result[fld.alias] = fld.decode(field)

        return cls(result)


@dataclass
class P4TableAction:
    """Represents a P4Runtime Action reference.

    e.g. P4TableAction("ipv4_forward", port=1)
    """

    name: str
    args: dict[str, Any]

    def __init__(self, __name: str, /, **args: Any):
        self.name = __name
        self.args = args

    def encode_table_action(self, table: P4Table) -> p4r.TableAction:
        "Encode TableAction data as protobuf."

        try:
            action = table.actions[self.name]
        except Exception as ex:
            raise ValueError(f"{table.name!r}: {ex}") from ex

        return p4r.TableAction(action=self._encode_action(action))

    def _fail_missing_params(self, action: P4ActionRef | P4Action) -> NoReturn:
        "Report missing parameters."

        seen = {param.name for param in action.params}
        for name in self.args:
            param = action.params[name]
            seen.remove(param.name)

        raise ValueError(f"{action.alias!r}: missing parameters {seen}")

    def encode_action(self, schema: P4Schema) -> p4r.Action:
        "Encode Action data as protobuf."

        # TODO: Make sure that action is normal `Action`.
        action = schema.actions[self.name]
        return self._encode_action(action)

    def _encode_action(self, action: P4ActionRef | P4Action) -> p4r.Action:
        "Helper to encode an action."

        params: list[p4r.Action.Param] = []
        for name, value in self.args.items():
            try:
                param = action.params[name]
                params.append(param.encode(value))
            except ValueError as ex:
                raise ValueError(f"{action.alias!r}: {ex}") from ex

        # Check for missing action parameters.
        if len(params) != len(action.params):
            self._fail_missing_params(action)

        return p4r.Action(action_id=action.id, params=params)

    @classmethod
    def decode_table_action(cls, msg: p4r.TableAction, table: P4Table) -> Self:
        "Decode protobuf to TableAction data."

        match msg.WhichOneof("type"):
            case "action":
                return cls.decode_action(msg.action, table)
            case "action_profile_member_id":
                raise NotImplementedError()
            case "action_profile_group_id":
                raise NotImplementedError()
            case "action_profile_action_set":
                raise NotImplementedError()
            case other:
                raise ValueError(f"unknown oneof: {other!r}")

    @classmethod
    def decode_action(cls, msg: p4r.Action, parent: P4Schema | P4Table) -> Self:
        "Decode protobuf to Action data."

        action = parent.actions[msg.action_id]
        args = {}
        for param in msg.params:
            action_param = action.params[param.param_id]
            value = action_param.decode(param)
            args[action_param.name] = value

        return cls(action.alias, **args)


@dataclass(kw_only=True)
class P4MeterConfig:
    "Represents a P4Runtime MeterConfig."

    cir: int
    cburst: int
    pir: int
    pburst: int

    def encode(self) -> p4r.MeterConfig:
        return p4r.MeterConfig(**self.__dict__)

    @classmethod
    def decode(cls, msg: p4r.MeterConfig) -> Self:
        return cls(cir=msg.cir, cburst=msg.cburst, pir=msg.pir, pburst=msg.pburst)


@dataclass(kw_only=True)
class P4CounterData:
    "Represents a P4Runtime CounterData."

    byte_count: int
    packet_count: int

    def encode(self) -> p4r.CounterData:
        return p4r.CounterData(**self.__dict__)

    @classmethod
    def decode(cls, msg: p4r.CounterData) -> Self:
        return cls(byte_count=msg.byte_count, packet_count=msg.packet_count)


@dataclass(kw_only=True)
class P4MeterCounterData:
    "Represents a P4Runtime MeterCounterData."

    green: P4CounterData
    yellow: P4CounterData
    red: P4CounterData

    def encode(self) -> p4r.MeterCounterData:
        return p4r.MeterCounterData(
            green=self.green.encode(),
            yellow=self.yellow.encode(),
            red=self.red.encode(),
        )

    @classmethod
    def decode(cls, msg: p4r.MeterCounterData) -> Self:
        return cls(
            green=P4CounterData.decode(msg.green),
            yellow=P4CounterData.decode(msg.yellow),
            red=P4CounterData.decode(msg.red),
        )


@decodable("table_entry")
@dataclass
class P4TableEntry(_Writable):
    "Represents a P4Runtime TableEntry."

    table_id: str = ""
    _: KW_ONLY
    match: P4TableMatch | None = None
    action: P4TableAction | None = None
    priority: int = 0
    meter_config: P4MeterConfig | None = None
    counter_data: P4CounterData | None = None
    meter_counter_data: P4MeterCounterData | None = None
    is_default_action: bool = False
    idle_timeout_ns: int = 0
    time_since_last_hit: int | None = None
    metadata: bytes = b""

    def encode(self, schema: P4Schema) -> p4r.Entity:
        "Encode TableEntry data as protobuf."

        return p4r.Entity(table_entry=self.encode_entry(schema))

    def encode_entry(self, schema: P4Schema) -> p4r.TableEntry:
        "Encode TableEntry data as protobuf."

        if not self.table_id:
            return p4r.TableEntry()

        table = schema.tables[self.table_id]

        if self.match:
            match = self.match.encode(table)
        else:
            match = None

        if self.action:
            action = self.action.encode_table_action(table)
        else:
            action = None

        if self.meter_config:
            meter_config = self.meter_config.encode()
        else:
            meter_config = None

        if self.counter_data:
            counter_data = self.counter_data.encode()
        else:
            counter_data = None

        if self.meter_counter_data:
            meter_counter_data = self.meter_counter_data.encode()
        else:
            meter_counter_data = None

        if self.time_since_last_hit is not None:
            time_since_last_hit = p4r.TableEntry.IdleTimeout(
                elapsed_ns=self.time_since_last_hit
            )
        else:
            time_since_last_hit = None

        return p4r.TableEntry(
            table_id=table.id,
            match=match,
            action=action,
            priority=self.priority,
            meter_config=meter_config,
            counter_data=counter_data,
            meter_counter_data=meter_counter_data,
            is_default_action=self.is_default_action,
            idle_timeout_ns=self.idle_timeout_ns,
            time_since_last_hit=time_since_last_hit,
            metadata=self.metadata,
        )

    @classmethod
    def decode(cls, msg: p4r.Entity, schema: P4Schema) -> Self:
        "Decode protobuf to TableEntry data."

        return cls.decode_entry(msg.table_entry, schema)

    @classmethod
    def decode_entry(cls, entry: p4r.TableEntry, schema: P4Schema) -> Self:
        "Decode protobuf to TableEntry data."

        if entry.table_id == 0:
            return cls("")

        table = schema.tables[entry.table_id]

        if entry.match:
            match = P4TableMatch.decode(entry.match, table)
        else:
            match = None

        if entry.HasField("action"):
            action = P4TableAction.decode_table_action(entry.action, table)
        else:
            action = None

        if entry.HasField("time_since_last_hit"):
            last_hit = entry.time_since_last_hit.elapsed_ns
        else:
            last_hit = None

        if entry.HasField("meter_config"):
            meter_config = P4MeterConfig.decode(entry.meter_config)
        else:
            meter_config = None

        # TODO: More fields.

        return cls(
            table_id=table.alias,
            match=match,
            action=action,
            priority=entry.priority,
            meter_config=meter_config,
            is_default_action=entry.is_default_action,
            idle_timeout_ns=entry.idle_timeout_ns,
            time_since_last_hit=last_hit,
            metadata=entry.metadata,
        )


@decodable("register_entry")
@dataclass
class P4RegisterEntry(_Writable):
    "Represents a P4Runtime RegisterEntry."

    register_id: str = ""
    _: KW_ONLY
    index: int | None = None
    data: _DataValueType | None = None

    def encode(self, schema: P4Schema) -> p4r.Entity:
        "Encode RegisterEntry data as protobuf."

        if not self.register_id:
            return p4r.Entity(register_entry=p4r.RegisterEntry())

        register = schema.registers[self.register_id]

        if self.index is not None:
            index = p4r.Index(index=self.index)
        else:
            index = None

        if self.data is not None:
            data = register.type_spec.encode_data(self.data)
        else:
            data = None

        entry = p4r.RegisterEntry(
            register_id=register.id,
            index=index,
            data=data,
        )
        return p4r.Entity(register_entry=entry)

    @classmethod
    def decode(cls, msg: p4r.Entity, schema: P4Schema) -> Self:
        "Decode protobuf to RegisterEntry data."

        entry = msg.register_entry
        if entry.register_id == 0:
            return cls("")

        register = schema.registers[entry.register_id]

        if entry.HasField("index"):
            index = entry.index.index
        else:
            index = None

        if entry.HasField("data"):
            data = register.type_spec.decode_data(entry.data)
        else:
            data = None

        return cls(
            register.alias,
            index=index,
            data=data,
        )


@decodable("multicast_group_entry")
@dataclass
class P4MulticastGroupEntry(_Writable):
    "Represents a P4Runtime MulticastGroupEntry."

    multicast_group_id: int = 0
    _: KW_ONLY
    replicas: Sequence[_ReplicaType] = ()

    def encode(self, _schema: P4Schema) -> p4r.Entity:
        "Encode MulticastGroupEntry data as protobuf."

        entry = p4r.MulticastGroupEntry(
            multicast_group_id=self.multicast_group_id,
            replicas=[encode_replica(replica) for replica in self.replicas],
        )
        return p4r.Entity(
            packet_replication_engine_entry=p4r.PacketReplicationEngineEntry(
                multicast_group_entry=entry
            )
        )

    @classmethod
    def decode(cls, msg: p4r.Entity, _schema: P4Schema) -> Self:
        "Decode protobuf to MulticastGroupEntry data."

        entry = msg.packet_replication_engine_entry.multicast_group_entry
        return cls(
            multicast_group_id=entry.multicast_group_id,
            replicas=tuple(decode_replica(replica) for replica in entry.replicas),
        )


@decodable("clone_session_entry")
@dataclass
class P4CloneSessionEntry(_Writable):
    "Represents a P4Runtime CloneSessionEntry."

    session_id: int = 0
    _: KW_ONLY
    class_of_service: int = 0
    packet_length_bytes: int = 0
    replicas: Sequence[_ReplicaType] = ()

    def encode(self, _schema: P4Schema) -> p4r.Entity:
        "Encode CloneSessionEntry data as protobuf."

        entry = p4r.CloneSessionEntry(
            session_id=self.session_id,
            class_of_service=self.class_of_service,
            packet_length_bytes=self.packet_length_bytes,
            replicas=[encode_replica(replica) for replica in self.replicas],
        )
        return p4r.Entity(
            packet_replication_engine_entry=p4r.PacketReplicationEngineEntry(
                clone_session_entry=entry
            )
        )

    @classmethod
    def decode(cls, msg: p4r.Entity, _schema: P4Schema) -> Self:
        "Decode protobuf to CloneSessionEntry data."

        entry = msg.packet_replication_engine_entry.clone_session_entry
        return cls(
            session_id=entry.session_id,
            class_of_service=entry.class_of_service,
            packet_length_bytes=entry.packet_length_bytes,
            replicas=tuple(decode_replica(replica) for replica in entry.replicas),
        )


@decodable("digest_entry")
@dataclass
class P4DigestEntry(_Writable):
    "Represents a P4Runtime DigestEntry."

    digest_id: str
    _: KW_ONLY
    max_list_size: int = 0
    max_timeout_ns: int = 0
    ack_timeout_ns: int = 0

    def encode(self, schema: P4Schema) -> p4r.Entity:
        "Encode DigestEntry data as protobuf."
        digest = schema.digests[self.digest_id]

        if self.max_list_size == self.max_timeout_ns == self.ack_timeout_ns == 0:
            config = None
        else:
            config = p4r.DigestEntry.Config(
                max_timeout_ns=self.max_timeout_ns,
                max_list_size=self.max_list_size,
                ack_timeout_ns=self.ack_timeout_ns,
            )

        entry = p4r.DigestEntry(digest_id=digest.id, config=config)
        return p4r.Entity(digest_entry=entry)

    @classmethod
    def decode(cls, msg: p4r.Entity, schema: P4Schema) -> Self:
        "Decode protobuf to DigestEntry data."
        entry = msg.digest_entry
        digest = schema.digests[entry.digest_id]

        config = entry.config
        return cls(
            digest.alias,
            max_list_size=config.max_list_size,
            max_timeout_ns=config.max_timeout_ns,
            ack_timeout_ns=config.ack_timeout_ns,
        )


@decodable("action_profile_member")
@dataclass
class P4ActionProfileMember(_Writable):
    "Represents a P4Runtime ActionProfileMember."

    action_profile_id: int = 0
    _: KW_ONLY
    member_id: int = 0
    action: P4TableAction | None = None

    def encode(self, schema: P4Schema) -> p4r.Entity:
        "Encode P4ActionProfileMember as protobuf."

        if self.action:
            action = self.action.encode_action(schema)
        else:
            action = None

        entry = p4r.ActionProfileMember(
            action_profile_id=self.action_profile_id,
            member_id=self.member_id,
            action=action,
        )
        return p4r.Entity(action_profile_member=entry)

    @classmethod
    def decode(cls, msg: p4r.Entity, schema: P4Schema) -> Self:
        "Decode protobuf to ActionProfileMember data."
        entry = msg.action_profile_member

        if entry.HasField("action"):
            action = P4TableAction.decode_action(entry.action, schema)
        else:
            action = None

        return cls(
            action_profile_id=entry.action_profile_id,
            member_id=entry.member_id,
            action=action,
        )


@dataclass(kw_only=True)
class P4Member:
    "Represents an ActionProfileGroup Member."

    member_id: int
    weight: _Weight

    def encode(self) -> p4r.ActionProfileGroup.Member:
        "Encode P4Member as protobuf."

        watch = None
        watch_port = None

        match self.weight:
            case int(weight):
                pass
            case (int(weight), int(watch)):
                pass
            case (int(weight), bytes(watch_port)):
                pass
            case other:
                raise ValueError(f"unexpected weight: {other!r}")

        member = p4r.ActionProfileGroup.Member(
            member_id=self.member_id,
            weight=weight,
        )

        if watch is not None:
            member.watch = watch
        elif watch_port is not None:
            member.watch_port = watch_port

        return member

    @classmethod
    def decode(cls, msg: p4r.ActionProfileGroup.Member) -> Self:
        "Decode protobuf to P4Member."

        match msg.WhichOneof("watch_kind"):
            case "watch":
                weight = (msg.weight, msg.watch)
            case "watch_port":
                weight = (msg.weight, msg.watch_port)
            case other:
                raise ValueError(f"unknown oneof: {other!r}")

        return cls(member_id=msg.member_id, weight=weight)


@decodable("action_profile_group")
@dataclass
class P4ActionProfileGroup(_Writable):
    "Represents a P4Runtime ActionProfileGroup."

    action_profile_id: int = 0
    _: KW_ONLY
    group_id: int = 0
    max_size: int = 0
    members: Sequence[P4Member] | None = None

    def encode(self, _schema: P4Schema) -> p4r.Entity:
        "Encode P4ActionProfileGroup as protobuf."

        if self.members is not None:
            members = [member.encode() for member in self.members]
        else:
            members = None

        entry = p4r.ActionProfileGroup(
            action_profile_id=self.action_profile_id,
            group_id=self.group_id,
            members=members,
            max_size=self.max_size,
        )
        return p4r.Entity(action_profile_group=entry)

    @classmethod
    def decode(cls, msg: p4r.Entity, _schema: P4Schema) -> Self:
        "Decode protobuf to ActionProfileGroup data."
        entry = msg.action_profile_group

        if entry.members:
            members = [P4Member.decode(member) for member in entry.members]
        else:
            members = None

        return cls(
            action_profile_id=entry.action_profile_id,
            group_id=entry.group_id,
            max_size=entry.max_size,
            members=members,
        )


@decodable("meter_entry")
@dataclass
class P4MeterEntry(_Writable):
    "Represents a P4Runtime MeterEntry."

    meter_id: int = 0
    _: KW_ONLY
    index: int | None = None
    config: P4MeterConfig | None = None
    counter_data: P4MeterCounterData | None = None

    def encode(self, _schema: P4Schema) -> p4r.Entity:
        "Encode P4MeterEntry to protobuf."

        if self.index is not None:
            index = p4r.Index(index=self.index)
        else:
            index = None

        if self.config is not None:
            config = self.config.encode()
        else:
            config = None

        if self.counter_data is not None:
            counter_data = self.counter_data.encode()
        else:
            counter_data = None

        entry = p4r.MeterEntry(
            meter_id=self.meter_id,
            index=index,
            config=config,
            counter_data=counter_data,
        )
        return p4r.Entity(meter_entry=entry)

    @classmethod
    def decode(cls, msg: p4r.Entity, _schema: P4Schema) -> Self:
        "Decode protobuf to P4MeterEntry."

        entry = msg.meter_entry

        if entry.HasField("index"):
            index = entry.index.index
        else:
            index = None

        if entry.HasField("config"):
            config = P4MeterConfig.decode(entry.config)
        else:
            config = None

        if entry.HasField("counter_data"):
            counter_data = P4MeterCounterData.decode(entry.counter_data)
        else:
            counter_data = None

        return cls(
            meter_id=entry.meter_id,
            index=index,
            config=config,
            counter_data=counter_data,
        )


@decodable("direct_meter_entry")
@dataclass
class P4DirectMeterEntry(_Writable):
    "Represents a P4Runtime DirectMeterEntry."

    table_entry: P4TableEntry | None = None
    _: KW_ONLY
    config: P4MeterConfig | None = None
    counter_data: P4MeterCounterData | None = None

    def encode(self, schema: P4Schema) -> p4r.Entity:
        "Encode P4DirectMeterEntry as protobuf."

        if self.table_entry is not None:
            table_entry = self.table_entry.encode_entry(schema)
        else:
            table_entry = p4r.TableEntry()

        if self.config is not None:
            config = self.config.encode()
        else:
            config = None

        if self.counter_data is not None:
            counter_data = self.counter_data.encode()
        else:
            counter_data = None

        entry = p4r.DirectMeterEntry(
            table_entry=table_entry,
            config=config,
            counter_data=counter_data,
        )
        return p4r.Entity(direct_meter_entry=entry)

    @classmethod
    def decode(cls, msg: p4r.Entity, schema: P4Schema) -> Self:
        "Decode protobuf to P4DirectMeterEntry."

        entry = msg.direct_meter_entry
        table_entry = P4TableEntry.decode_entry(entry.table_entry, schema)

        if entry.HasField("config"):
            config = P4MeterConfig.decode(entry.config)
        else:
            config = None

        if entry.HasField("counter_data"):
            counter_data = P4MeterCounterData.decode(entry.counter_data)
        else:
            counter_data = None

        return cls(
            table_entry=table_entry,
            config=config,
            counter_data=counter_data,
        )


@decodable("counter_entry")
@dataclass
class P4CounterEntry(_Writable):
    "Represents a P4Runtime CounterEntry."

    counter_id: int = 0
    _: KW_ONLY
    index: int | None = None
    data: P4CounterData | None = None

    def encode(self, _schema: P4Schema) -> p4r.Entity:
        "Encode P4CounterEntry as protobuf."

        if self.index is not None:
            index = p4r.Index(index=self.index)
        else:
            index = None

        if self.data is not None:
            data = self.data.encode()
        else:
            data = None

        entry = p4r.CounterEntry(
            counter_id=self.counter_id,
            index=index,
            data=data,
        )
        return p4r.Entity(counter_entry=entry)

    @classmethod
    def decode(cls, msg: p4r.Entity, _schema: P4Schema) -> Self:
        "Decode protobuf to P4CounterEntry."

        entry = msg.counter_entry

        if entry.HasField("index"):
            index = entry.index.index
        else:
            index = None

        if entry.HasField("data"):
            data = P4CounterData.decode(entry.data)
        else:
            data = None

        return cls(counter_id=entry.counter_id, index=index, data=data)


@decodable("direct_counter_entry")
@dataclass
class P4DirectCounterEntry(_Writable):
    "Represents a P4Runtime DirectCounterEntry."

    table_entry: P4TableEntry | None = None
    _: KW_ONLY
    data: P4CounterData | None = None

    def encode(self, schema: P4Schema) -> p4r.Entity:
        "Encode P4DirectCounterEntry as protobuf."

        if self.table_entry is not None:
            table_entry = self.table_entry.encode_entry(schema)
        else:
            table_entry = None

        if self.data is not None:
            data = self.data.encode()
        else:
            data = None

        entry = p4r.DirectCounterEntry(
            table_entry=table_entry,
            data=data,
        )
        return p4r.Entity(direct_counter_entry=entry)

    @classmethod
    def decode(cls, msg: p4r.Entity, schema: P4Schema) -> Self:
        "Decode protobuf to P4DirectCounterEntry."

        entry = msg.direct_counter_entry

        if entry.HasField("table_entry"):
            table_entry = P4TableEntry.decode_entry(entry.table_entry, schema)
        else:
            table_entry = None

        if entry.HasField("data"):
            data = P4CounterData.decode(entry.data)
        else:
            data = None

        return cls(table_entry=table_entry, data=data)


@decodable("packet")
@dataclass
class P4PacketIn:
    "Represents a P4Runtime PacketIn."

    payload: bytes
    _: KW_ONLY
    metadata: _MetadataDictType

    @classmethod
    def decode(cls, msg: p4r.StreamMessageResponse, schema: P4Schema) -> Self:
        "Decode protobuf to PacketIn data."

        packet = msg.packet
        cpm = schema.controller_packet_metadata.get("packet_in")
        if cpm is None:
            # There is no controller metadata. Warn if message has any.
            if packet.HasField("metadata"):
                LOGGER.warning("unexpected metadata: %r", packet.metadata)
            return cls(packet.payload, metadata={})

        return cls(
            packet.payload,
            metadata=cpm.decode(packet.metadata),
        )

    def __getitem__(self, key: str):
        "Retrieve metadata value."

        return self.metadata[key]

    def __repr__(self):
        "Return friendlier hexadecimal description of packet."

        if self.metadata:
            return (
                f"PacketIn(metadata={self.metadata!r}, payload=h'{self.payload.hex()}')"
            )
        return f"PacketIn(payload=h'{self.payload.hex()}')"


@dataclass
class P4PacketOut:
    "Represents a P4Runtime PacketOut."

    payload: bytes
    metadata: _MetadataDictType

    def __init__(self, __payload: bytes, /, **metadata: Any):
        self.payload = __payload
        self.metadata = metadata

    def encode_update(self, schema: P4Schema) -> p4r.StreamMessageRequest:
        "Encode PacketOut data as protobuf."

        cpm = schema.controller_packet_metadata["packet_out"]
        return p4r.StreamMessageRequest(
            packet=p4r.PacketOut(
                payload=self.payload,
                metadata=cpm.encode(self.metadata),
            )
        )

    def __getitem__(self, key: str):
        "Retrieve metadata value."

        return self.metadata[key]

    def __repr__(self):
        "Return friendlier hexadecimal description of packet."

        if self.metadata:
            return f"PacketOut(metadata={self.metadata!r}, payload=h'{self.payload.hex()}')"
        return f"PacketOut(payload=h'{self.payload.hex()}')"


@decodable("digest")
@dataclass
class P4DigestList:
    "Represents a P4Runtime DigestList."

    digest_id: str
    _: KW_ONLY
    list_id: int
    timestamp: int
    data: list[_DataValueType]

    @classmethod
    def decode(cls, msg: p4r.StreamMessageResponse, schema: P4Schema) -> Self:
        "Decode protobuf to DigestList data."

        digest_list = msg.digest
        digest = schema.digests[digest_list.digest_id]

        type_spec = digest.type_spec
        return cls(
            digest_id=digest.alias,
            list_id=digest_list.list_id,
            timestamp=digest_list.timestamp,
            data=[type_spec.decode_data(item) for item in digest_list.data],
        )

    def __len__(self):
        "Return number of values in digest list."
        return len(self.data)

    def __getitem__(self, key: int):
        "Retrieve value at given index from digest list."
        return self.data[key]

    def __iter__(self):
        "Iterate over values in digest list."
        return iter(self.data)


@dataclass
class P4DigestListAck:
    "Represents a P4Runtime DigestListAck."


@decodable("idle_timeout_notification")
@dataclass
class P4IdleTimeoutNotification:
    "Represents a P4Runtime IdleTimeoutNotification."



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.enum_type_wrapper
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _Code:
    ValueType = typing.NewType('ValueType', builtins.int)
    V: typing_extensions.TypeAlias = ValueType
class _CodeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_Code.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    OK: _Code.ValueType  # 0
    """Not an error; returned on success

    HTTP Mapping: 200 OK
    """

    CANCELLED: _Code.ValueType  # 1
    """The operation was cancelled, typically by the caller.

    HTTP Mapping: 499 Client Closed Request
    """

    UNKNOWN: _Code.ValueType  # 2
    """Unknown error.  For example, this error may be returned when
    a `Status` value received from another address space belongs to
    an error space that is not known in this address space.  Also
    errors raised by APIs that do not return enough error information
    may be converted to this error.

    HTTP Mapping: 500 Internal Server Error
    """

    INVALID_ARGUMENT: _Code.ValueType  # 3
    """The client specified an invalid argument.  Note that this differs
    from `FAILED_PRECONDITION`.  `INVALID_ARGUMENT` indicates arguments
    that are problematic regardless of the state of the system
    (e.g., a malformed file name).

    HTTP Mapping: 400 Bad Request
    """

    DEADLINE_EXCEEDED: _Code.ValueType  # 4
    """The deadline expired before the operation could complete. For operations
    that change the state of the system, this error may be returned
    even if the operation has completed successfully.  For example, a
    successful response from a server could have been delayed long
    enough for the deadline to expire.

    HTTP Mapping: 504 Gateway Timeout
    """

    NOT_FOUND: _Code.ValueType  # 5
    """Some requested entity (e.g., file or directory) was not found.

    Note to server developers: if a request is denied for an entire class
    of users, such as gradual feature rollout or undocumented whitelist,
    `NOT_FOUND` may be used. If a request is denied for some users within
    a class of users, such as user-based access control, `PERMISSION_DENIED`
    must be used.

    HTTP Mapping: 404 Not Found
    """

    ALREADY_EXISTS: _Code.ValueType  # 6
    """The entity that a client attempted to create (e.g., file or directory)
    already exists.

    HTTP Mapping: 409 Conflict
    """

    PERMISSION_DENIED: _Code.ValueType  # 7
    """The caller does not have permission to execute the specified
    operation. `PERMISSION_DENIED` must not be used for rejections
    caused by exhausting some resource (use `RESOURCE_EXHAUSTED`
    instead for those errors). `PERMISSION_DENIED` must not be
    used if the caller can not be identified (use `UNAUTHENTICATED`
    instead for those errors). This error code does not imply the
    request is valid or the requested entity exists or satisfies
    other pre-conditions.

    HTTP Mapping: 403 Forbidden
    """

    UNAUTHENTICATED: _Code.ValueType  # 16
    """The request does not have valid authentication credentials for the
    operation.

    HTTP Mapping: 401 Unauthorized
    """

    RESOURCE_EXHAUSTED: _Code.ValueType  # 8
    """Some resource has been exhausted, perhaps a per-user quota, or
    perhaps the entire file system is out of space.

    HTTP Mapping: 429 Too Many Requests
    """

    FAILED_PRECONDITION: _Code.ValueType  # 9
    """The operation was rejected because the system is not in a state
    required for the operation's execution.  For example, the directory
    to be deleted is non-empty, an rmdir operation is applied to
    a non-directory, etc.

    Service implementors can use the following guidelines to decide
    between `FAILED_PRECONDITION`, `ABORTED`, and `UNAVAILABLE`:
     (a) Use `UNAVAILABLE` if the client can retry just the failing call.
     (b) Use `ABORTED` if the client should retry at a higher level
         (e.g., when a client-specified test-and-set fails, indicating the
         client should restart a read-modify-write sequence).
     (c) Use `FAILED_PRECONDITION` if the client should not retry until
         the system state has been explicitly fixed.  E.g., if an "rmdir"
         fails because the directory is non-empty, `FAILED_PRECONDITION`
         should be returned since the client should not retry unless
         the files are deleted from the directory.

    HTTP Mapping: 400 Bad Request
    """

    ABORTED: _Code.ValueType  # 10
    """The operation was aborted, typically due to a concurrency issue such as
    a sequencer check failure or transaction abort.

    See the guidelines above for deciding between `FAILED_PRECONDITION`,
    `ABORTED`, and `UNAVAILABLE`.

    HTTP Mapping: 409 Conflict
    """

    OUT_OF_RANGE: _Code.ValueType  # 11
    """The operation was attempted past the valid range.  E.g., seeking or
    reading past end-of-file.

    Unlike `INVALID_ARGUMENT`, this error indicates a problem that may
    be fixed if the system state changes. For example, a 32-bit file
    system will generate `INVALID_ARGUMENT` if asked to read at an
    offset that is not in the range [0,2^32-1], but it will generate
    `OUT_OF_RANGE` if asked to read from an offset past the current
    file size.

    There is a fair bit of overlap between `FAILED_PRECONDITION` and
    `OUT_OF_RANGE`.  We recommend using `OUT_OF_RANGE` (the more specific
    error) when it applies so that callers who are iterating through
    a space can easily look for an `OUT_OF_RANGE` error to detect when
    they are done.

    HTTP Mapping: 400 Bad Request
    """

    UNIMPLEMENTED: _Code.ValueType  # 12
    """The operation is not implemented or is not supported/enabled in this
    service.

    HTTP Mapping: 501 Not Implemented
    """

    INTERNAL: _Code.ValueType  # 13
    """Internal errors.  This means that some invariants expected by the
    underlying system have been broken.  This error code is reserved
    for serious errors.

    HTTP Mapping: 500 Internal Server Error
    """

    UNAVAILABLE: _Code.ValueType  # 14
    """The service is currently unavailable.  This is most likely a
    transient condition, which can be corrected by retrying with
    a backoff. Note that it is not always safe to retry
    non-idempotent operations.

    See the guidelines above for deciding between `FAILED_PRECONDITION`,
    `ABORTED`, and `UNAVAILABLE`.

    HTTP Mapping: 503 Service Unavailable
    """

    DATA_LOSS: _Code.ValueType  # 15
    """Unrecoverable data loss or corruption.

    HTTP Mapping: 500 Internal Server Error
    """

class Code(_Code, metaclass=_CodeEnumTypeWrapper):
    """The canonical error codes for gRPC APIs.


    Sometimes multiple error codes may apply.  Services should return
    the most specific error code that applies.  For example, prefer
    `OUT_OF_RANGE` over `FAILED_PRECONDITION` if both codes apply.
    Similarly prefer `NOT_FOUND` or `ALREADY_EXISTS` over `FAILED_PRECONDITION`.
    """
    pass

OK: Code.ValueType  # 0
"""Not an error; returned on success

HTTP Mapping: 200 OK
"""

CANCELLED: Code.ValueType  # 1
"""The operation was cancelled, typically by the caller.

HTTP Mapping: 499 Client Closed Request
"""

UNKNOWN: Code.ValueType  # 2
"""Unknown error.  For example, this error may be returned when
a `Status` value received from another address space belongs to
an error space that is not known in this address space.  Also
errors raised by APIs that do not return enough error information
may be converted to this error.

HTTP Mapping: 500 Internal Server Error
"""

INVALID_ARGUMENT: Code.ValueType  # 3
"""The client specified an invalid argument.  Note that this differs
from `FAILED_PRECONDITION`.  `INVALID_ARGUMENT` indicates arguments
that are problematic regardless of the state of the system
(e.g., a malformed file name).

HTTP Mapping: 400 Bad Request
"""

DEADLINE_EXCEEDED: Code.ValueType  # 4
"""The deadline expired before the operation could complete. For operations
that change the state of the system, this error may be returned
even if the operation has completed successfully.  For example, a
successful response from a server could have been delayed long
enough for the deadline to expire.

HTTP Mapping: 504 Gateway Timeout
"""

NOT_FOUND: Code.ValueType  # 5
"""Some requested entity (e.g., file or directory) was not found.

Note to server developers: if a request is denied for an entire class
of users, such as gradual feature rollout or undocumented whitelist,
`NOT_FOUND` may be used. If a request is denied for some users within
a class of users, such as user-based access control, `PERMISSION_DENIED`
must be used.

HTTP Mapping: 404 Not Found
"""

ALREADY_EXISTS: Code.ValueType  # 6
"""The entity that a client attempted to create (e.g., file or directory)
already exists.

HTTP Mapping: 409 Conflict
"""

PERMISSION_DENIED: Code.ValueType  # 7
"""The caller does not have permission to execute the specified
operation. `PERMISSION_DENIED` must not be used for rejections
caused by exhausting some resource (use `RESOURCE_EXHAUSTED`
instead for those errors). `PERMISSION_DENIED` must not be
used if the caller can not be identified (use `UNAUTHENTICATED`
instead for those errors). This error code does not imply the
request is valid or the requested entity exists or satisfies
other pre-conditions.

HTTP Mapping: 403 Forbidden
"""

UNAUTHENTICATED: Code.ValueType  # 16
"""The request does not have valid authentication credentials for the
operation.

HTTP Mapping: 401 Unauthorized
"""

RESOURCE_EXHAUSTED: Code.ValueType  # 8
"""Some resource has been exhausted, perhaps a per-user quota, or
perhaps the entire file system is out of space.

HTTP Mapping: 429 Too Many Requests
"""

FAILED_PRECONDITION: Code.ValueType  # 9
"""The operation was rejected because the system is not in a state
required for the operation's execution.  For example, the directory
to be deleted is non-empty, an rmdir operation is applied to
a non-directory, etc.

Service implementors can use the following guidelines to decide
between `FAILED_PRECONDITION`, `ABORTED`, and `UNAVAILABLE`:
 (a) Use `UNAVAILABLE` if the client can retry just the failing call.
 (b) Use `ABORTED` if the client should retry at a higher level
     (e.g., when a client-specified test-and-set fails, indicating the
     client should restart a read-modify-write sequence).
 (c) Use `FAILED_PRECONDITION` if the client should not retry until
     the system state has been explicitly fixed.  E.g., if an "rmdir"
     fails because the directory is non-empty, `FAILED_PRECONDITION`
     should be returned since the client should not retry unless
     the files are deleted from the directory.

HTTP Mapping: 400 Bad Request
"""

ABORTED: Code.ValueType  # 10
"""The operation was aborted, typically due to a concurrency issue such as
a sequencer check failure or transaction abort.

See the guidelines above for deciding between `FAILED_PRECONDITION`,
`ABORTED`, and `UNAVAILABLE`.

HTTP Mapping: 409 Conflict
"""

OUT_OF_RANGE: Code.ValueType  # 11
"""The operation was attempted past the valid range.  E.g., seeking or
reading past end-of-file.

Unlike `INVALID_ARGUMENT`, this error indicates a problem that may
be fixed if the system state changes. For example, a 32-bit file
system will generate `INVALID_ARGUMENT` if asked to read at an
offset that is not in the range [0,2^32-1], but it will generate
`OUT_OF_RANGE` if asked to read from an offset past the current
file size.

There is a fair bit of overlap between `FAILED_PRECONDITION` and
`OUT_OF_RANGE`.  We recommend using `OUT_OF_RANGE` (the more specific
error) when it applies so that callers who are iterating through
a space can easily look for an `OUT_OF_RANGE` error to detect when
they are done.

HTTP Mapping: 400 Bad Request
"""

UNIMPLEMENTED: Code.ValueType  # 12
"""The operation is not implemented or is not supported/enabled in this
service.

HTTP Mapping: 501 Not Implemented
"""

INTERNAL: Code.ValueType  # 13
"""Internal errors.  This means that some invariants expected by the
underlying system have been broken.  This error code is reserved
for serious errors.

HTTP Mapping: 500 Internal Server Error
"""

UNAVAILABLE: Code.ValueType  # 14
"""The service is currently unavailable.  This is most likely a
transient condition, which can be corrected by retrying with
a backoff. Note that it is not always safe to retry
non-idempotent operations.

See the guidelines above for deciding between `FAILED_PRECONDITION`,
`ABORTED`, and `UNAVAILABLE`.

HTTP Mapping: 503 Service Unavailable
"""

DATA_LOSS: Code.ValueType  # 15
"""Unrecoverable data loss or corruption.

HTTP Mapping: 500 Internal Server Error
"""

global___Code = Code




# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: gnmi1/gnmi_ext.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14gnmi1/gnmi_ext.proto\x12\x08gnmi_ext\"\xac\x01\n\tExtension\x12\x37\n\x0eregistered_ext\x18\x01 \x01(\x0b\x32\x1d.gnmi_ext.RegisteredExtensionH\x00\x12\x39\n\x12master_arbitration\x18\x02 \x01(\x0b\x32\x1b.gnmi_ext.MasterArbitrationH\x00\x12$\n\x07history\x18\x03 \x01(\x0b\x32\x11.gnmi_ext.HistoryH\x00\x42\x05\n\x03\x65xt\"E\n\x13RegisteredExtension\x12!\n\x02id\x18\x01 \x01(\x0e\x32\x15.gnmi_ext.ExtensionID\x12\x0b\n\x03msg\x18\x02 \x01(\x0c\"Y\n\x11MasterArbitration\x12\x1c\n\x04role\x18\x01 \x01(\x0b\x32\x0e.gnmi_ext.Role\x12&\n\x0b\x65lection_id\x18\x02 \x01(\x0b\x32\x11.gnmi_ext.Uint128\"$\n\x07Uint128\x12\x0c\n\x04high\x18\x01 \x01(\x04\x12\x0b\n\x03low\x18\x02 \x01(\x04\"\x12\n\x04Role\x12\n\n\x02id\x18\x01 \x01(\t\"S\n\x07History\x12\x17\n\rsnapshot_time\x18\x01 \x01(\x03H\x00\x12$\n\x05range\x18\x02 \x01(\x0b\x32\x13.gnmi_ext.TimeRangeH\x00\x42\t\n\x07request\"\'\n\tTimeRange\x12\r\n\x05start\x18\x01 \x01(\x03\x12\x0b\n\x03\x65nd\x18\x02 \x01(\x03*3\n\x0b\x45xtensionID\x12\r\n\tEID_UNSET\x10\x00\x12\x15\n\x10\x45ID_EXPERIMENTAL\x10\xe7\x07\x42+Z)github.com/openconfig/gnmi/proto/gnmi_extb\x06proto3')

_EXTENSIONID = DESCRIPTOR.enum_types_by_name['ExtensionID']
ExtensionID = enum_type_wrapper.EnumTypeWrapper(_EXTENSIONID)
EID_UNSET = 0
EID_EXPERIMENTAL = 999


_EXTENSION = DESCRIPTOR.message_types_by_name['Extension']
_REGISTEREDEXTENSION = DESCRIPTOR.message_types_by_name['RegisteredExtension']
_MASTERARBITRATION = DESCRIPTOR.message_types_by_name['MasterArbitration']
_UINT128 = DESCRIPTOR.message_types_by_name['Uint128']
_ROLE = DESCRIPTOR.message_types_by_name['Role']
_HISTORY = DESCRIPTOR.message_types_by_name['History']
_TIMERANGE = DESCRIPTOR.message_types_by_name['TimeRange']
Extension = _reflection.GeneratedProtocolMessageType('Extension', (_message.Message,), {
  'DESCRIPTOR' : _EXTENSION,
  '__module__' : 'gnmi1.gnmi_ext_pb2'
  # @@protoc_insertion_point(class_scope:gnmi_ext.Extension)
  })
_sym_db.RegisterMessage(Extension)

RegisteredExtension = _reflection.GeneratedProtocolMessageType('RegisteredExtension', (_message.Message,), {
  'DESCRIPTOR' : _REGISTEREDEXTENSION,
  '__module__' : 'gnmi1.gnmi_ext_pb2'
  # @@protoc_insertion_point(class_scope:gnmi_ext.RegisteredExtension)
  })
_sym_db.RegisterMessage(RegisteredExtension)

MasterArbitration = _reflection.GeneratedProtocolMessageType('MasterArbitration', (_message.Message,), {
  'DESCRIPTOR' : _MASTERARBITRATION,
  '__module__' : 'gnmi1.gnmi_ext_pb2'
  # @@protoc_insertion_point(class_scope:gnmi_ext.MasterArbitration)
  })
_sym_db.RegisterMessage(MasterArbitration)

Uint128 = _reflection.GeneratedProtocolMessageType('Uint128', (_message.Message,), {
  'DESCRIPTOR' : _UINT128,
  '__module__' : 'gnmi1.gnmi_ext_pb2'
  # @@protoc_insertion_point(class_scope:gnmi_ext.Uint128)
  })
_sym_db.RegisterMessage(Uint128)

Role = _reflection.GeneratedProtocolMessageType('Role', (_message.Message,), {
  'DESCRIPTOR' : _ROLE,
  '__module__' : 'gnmi1.gnmi_ext_pb2'
  # @@protoc_insertion_point(class_scope:gnmi_ext.Role)
  })
_sym_db.RegisterMessage(Role)

History = _reflection.GeneratedProtocolMessageType('History', (_message.Message,), {
  'DESCRIPTOR' : _HISTORY,
  '__module__' : 'gnmi1.gnmi_ext_pb2'
  # @@protoc_insertion_point(class_scope:gnmi_ext.History)
  })
_sym_db.RegisterMessage(History)

TimeRange = _reflection.GeneratedProtocolMessageType('TimeRange', (_message.Message,), {
  'DESCRIPTOR' : _TIMERANGE,
  '__module__' : 'gnmi1.gnmi_ext_pb2'
  # @@protoc_insertion_point(class_scope:gnmi_ext.TimeRange)
  })
_sym_db.RegisterMessage(TimeRange)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z)github.com/openconfig/gnmi/proto/gnmi_ext'
  _EXTENSIONID._serialized_start=555
  _EXTENSIONID._serialized_end=606
  _EXTENSION._serialized_start=35
  _EXTENSION._serialized_end=207
  _REGISTEREDEXTENSION._serialized_start=209
  _REGISTEREDEXTENSION._serialized_end=278
  _MASTERARBITRATION._serialized_start=280
  _MASTERARBITRATION._serialized_end=369
  _UINT128._serialized_start=371
  _UINT128._serialized_end=407
  _ROLE._serialized_start=409
  _ROLE._serialized_end=427
  _HISTORY._serialized_start=429
  _HISTORY._serialized_end=512
  _TIMERANGE._serialized_start=514
  _TIMERANGE._serialized_end=553
# @@protoc_insertion_point(module_scope)



"Utility functions for protobuf messages."

# Copyright (c) 2022 Bill Fisher
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import TypeVar

import grpc  # pyright: ignore [reportMissingTypeStubs]
from google.protobuf import json_format, text_format
from google.protobuf.any_pb2 import Any as _Any  # pylint: disable=E0611
from google.protobuf.message import Message as _Message

import finsy as _fy
from finsy.gnmiclient import gNMIPath
from finsy.log import FINSY_TRANSLATE_LOGS, MSG_LOG
from finsy.proto import gnmi, p4r

PBAny = _Any
PBMessage = _Message

_MT = TypeVar("_MT", bound=PBMessage)


def from_any(any_obj: PBAny, msg_class: type[_MT]) -> _MT:
    "Unpack a protobuf `any_pb2.Any` message, or raise an exception."

    obj = msg_class()
    if not any_obj.Unpack(obj):  # pyright: ignore[reportUnknownMemberType]
        raise ValueError(f"Not a {msg_class.__name__}: {any_obj!r}")
    return obj


def from_text(data: str, msg_class: type[_MT]) -> _MT:
    "Read protobuf message from given text/json string."
    assert isinstance(data, str)

    msg = msg_class()
    if data[0] == "{":
        json_format.Parse(data, msg)
    else:
        text_format.Parse(data, msg)

    return msg


def from_dict(value: dict[str, int | str], msg_class: type[_MT]) -> _MT:
    "Convert Python dict to protobuf message."
    return json_format.ParseDict(value, msg_class())


def to_text(
    msg: PBMessage,
    *,
    as_one_line: bool = True,
    custom_format: bool = False,
) -> str:
    "Convert protobuf message to text format."
    assert isinstance(msg, PBMessage)
    formatter = _message_formatter if custom_format else None
    result = text_format.MessageToString(
        msg,
        as_one_line=as_one_line,
        message_formatter=formatter,
    )
    return result


def to_json(msg: PBMessage) -> str:
    "Convert protobuf message to JSON format."
    assert isinstance(msg, PBMessage), f"not a Message: {msg!r}"
    return json_format.MessageToJson(msg, preserving_proto_field_name=True)


def to_dict(msg: PBMessage) -> dict[str, int | str]:
    "Convert protobuf message to Python dict."
    assert isinstance(msg, PBMessage), f"not a Message: {msg!r}"
    return json_format.MessageToDict(msg, preserving_proto_field_name=True)


def _message_formatter(msg: PBMessage, _indent: int, _as_one_line: bool):
    if isinstance(msg, p4r.ForwardingPipelineConfig):
        return f"📦[p4cookie=0x{msg.cookie.cookie:x}]"
    if isinstance(msg, gnmi.Path):
        return f"📂{gNMIPath(msg)}"
    if isinstance(msg, gnmi.Update):
        value = repr(msg.val).strip()
        dups = "" if not msg.duplicates else f" ({msg.duplicates} dups)"
        return f"📂{gNMIPath(msg.path)} = {value}{dups}"
    if isinstance(msg, (p4r.PacketIn, p4r.PacketOut)):
        metadata = " ".join(
            f"meta[{md.metadata_id}]={md.value.hex()}" for md in msg.metadata
        )
        return f"📬{msg.payload.hex()} {metadata}"
    return None


def log_msg(
    state: grpc.ChannelConnectivity,
    msg: PBMessage,
    schema: "_fy.P4Schema | None",
):
    """Log a sent/received client message.

    Format:
        <state><mesg-type> (<size> bytes): <msg-contents>

    <state> is empty if the client state is READY. Otherwise, it's the
    channel connectivity state.
    """
    # Include the channel's state if it's not READY.
    state_name = ""
    if state != grpc.ChannelConnectivity.READY:
        state_name = f"{state.name} "  # trailing space necessary

    custom_format = FINSY_TRANSLATE_LOGS

    if isinstance(msg, (p4r.ReadResponse, p4r.WriteRequest)):
        text = to_text(msg, as_one_line=False, custom_format=custom_format)
        if schema is not None:
            text = _log_annotate(text, schema)
    elif isinstance(msg, gnmi.GetResponse):
        text = to_text(msg, as_one_line=False, custom_format=custom_format)
    else:
        text = to_text(msg, custom_format=custom_format)

    size = msg.ByteSize()

    MSG_LOG.debug(
        "%s%s (%d bytes): %s",
        state_name,
        type(msg).__name__,
        size,
        text,
    )


_ANNOTATE_REGEX = re.compile(r"([a-z]+_id): (\d+)\n", re.MULTILINE)


def _log_annotate(text: str, schema: "_fy.P4Schema") -> str:
    "Annotate table_id, action_id, etc. in log messages."

    action_id = 0
    table_id = 0

    def _replace(m: re.Match[str]) -> str:
        nonlocal action_id, table_id

        key, value = m.groups()
        name = None

        try:
            match key:
                case "table_id":
                    table_id = int(value)
                    name = schema.tables[table_id].name
                case "action_id":
                    action_id = int(value)
                    name = schema.actions[action_id].name
                case "digest_id":
                    name = schema.digests[int(value)].name
                case "field_id":
                    name = schema.tables[table_id].match_fields[int(value)].name
                case "param_id":
                    name = schema.actions[action_id].params[int(value)].name
                # TODO: Annotate more identifier types.
                case _:
                    pass

        except (LookupError, ValueError):
            # If there is a failure, don't replace anything.
            pass

        if name:
            return f"{key}: {value}  # {name}\n"

        return m[0]

    return _ANNOTATE_REGEX.sub(_replace, text)
